
import six
from math import copysign
import numpy as np
import mceq_config as config
from MCEq.misc import info, print_in_rows, getAZN

from particletools.tables import PYTHIAParticleData
info(5, 'Initialization of PYTHIAParticleData object')
_pdata = PYTHIAParticleData()

backward_compatible_namestr = {
    'nu_mu': 'numu',
    'nu_mubar': 'antinumu',
    'nu_e': 'nue',
    'nu_ebar': 'antinue',
    'nu_tau': 'nutau',
    'nu_taubar': 'antinutau'
}


# Replace particle names for neutrinos with those used
# in previous MCEq versions
def _pname(pdg_id_or_name):
    """Replace some particle names from pythia database with those from previous
    MCEq versions for backward compatibility."""

    pythia_name = _pdata.name(pdg_id_or_name)
    if pythia_name in backward_compatible_namestr:
        return backward_compatible_namestr[pythia_name]
    return pythia_name


class MCEqParticle(object):
    """Bundles different particle properties for simplified
    availability of particle properties in :class:`MCEq.core.MCEqRun`.

    Args:
      pdg_id (int): PDG ID of the particle
      egrid (np.array, optional): energy grid (centers)
      cs_db (object, optional): reference to an instance of
                                :class:`InteractionYields`
    """
    def __init__(self,
                 pdg_id,
                 helicity,
                 energy_grid=None,
                 cs_db=None,
                 init_pdata_defaults=True):

        #: (bool) if it's an electromagnetic particle
        self.is_em = abs(pdg_id) == 11 or pdg_id == 22
        #: (int) helicity -1, 0, 1 (0 means undefined or average)
        self.helicity = helicity
        #: (bool) particle is a nucleus (not yet implemented)
        self.is_nucleus = False
        #: (bool) particle is a hadron
        self.is_hadron = False
        #: (bool) particle is a hadron
        self.is_lepton = False
        #: (float) ctau in cm
        self.ctau = None
        #: (float) mass in GeV
        self.mass = None
        #: (str) species name in string representation
        self.name = None
        #: Mass, charge, neutron number
        self.A, self.Z, self.N = getAZN(pdg_id)
        #: (bool) particle has both, hadron and resonance properties
        self.is_mixed = False
        #: (bool) if particle has just resonance behavior
        self.is_resonance = False
        #: (bool) particle is interacting projectile
        self.is_projectile = False
        #: (bool) particle is stable
        self.is_stable = False or pdg_id in config.adv_set['disable_decays']
        #: (bool) can_interact
        self.can_interact = False
        #: (bool) has continuous losses dE/dX defined
        self.has_contloss = False
        #: (np.array) continuous losses in GeV/(g/cm2)
        self.dEdX = None
        #: (bool) is a tracking particle
        self.is_tracking = False
        #: decay channels if any
        self.decay_dists = {}
        #: (int) Particle Data Group Monte Carlo particle ID
        self.pdg_id = (pdg_id, helicity)
        #: (int) Unique PDG ID that is different for tracking particles
        self.unique_pdg_id = (pdg_id, helicity)
        #: (int) MCEq ID
        self.mceqidx = -1

        #: (float) mixing energy, transition between hadron and
        # resonance behavior
        self.E_mix = 0
        #: (int) energy grid index, where transition between
        # hadron and resonance occurs
        self.mix_idx = 0
        #: (float) critical energy in air at the surface
        self.E_crit = 0
        # Energy and cross section dependent inits
        self.current_cross_sections = None
        self._energy_grid = energy_grid

        # Variables for hadronic interaction
        self.current_hadronic_model = None
        self.hadr_secondaries = []
        self.hadr_yields = {}

        # Variables for decays
        self.children = []
        self.decay_dists = {}

        if init_pdata_defaults:
            self._init_defaults_from_pythia_database()

        if self._energy_grid is not None and cs_db is not None:
            #: interaction cross section in 1/cm2
            self.set_cs(cs_db)

    def _init_defaults_from_pythia_database(self):
        """Init some particle properties from :mod:`particletools.tables`."""
        #: (bool) particle is a nucleus (not yet implemented)
        self.is_nucleus = _pdata.is_nucleus(self.pdg_id[0])
        #: (bool) particle is a hadron
        self.is_hadron = _pdata.is_hadron(self.pdg_id[0])
        #: (bool) particle is a hadron
        self.is_lepton = _pdata.is_lepton(self.pdg_id[0])
        #: Mass, charge, neutron number
        self.A, self.Z, self.N = getAZN(self.pdg_id[0])
        #: (float) ctau in cm
        self.ctau = _pdata.ctau(self.pdg_id[0])
        #: (float) mass in GeV
        self.mass = _pdata.mass(self.pdg_id[0])
        #: (str) species name in string representation
        name = _pname(self.pdg_id[0]) if self.name is None else self.name
        if self.helicity == -1:
            name += '_l'
        elif self.helicity == +1:
            name += '_r'
        self.name = name
        #: (bool) particle is stable
        #: TODO the exclusion of neutron decays is a hotfix
        self.is_stable = (not self.ctau < np.inf or
                          self.pdg_id[0] in config.adv_set['disable_decays'])

    def init_custom_particle_data(self, name, pdg_id, helicity, ctau, mass,
                                  **kwargs):
        """Add custom particle type. (Incomplete and not debugged)"""
        #: (int) Particle Data Group Monte Carlo particle ID
        self.pdg_id = (pdg_id, helicity)
        #: (bool) if it's an electromagnetic particle
        self.is_em = kwargs.pop('is_em', abs(pdg_id) == 11 or pdg_id == 22)
        #: (bool) particle is a nucleus (not yet implemented)
        self.is_nucleus = kwargs.pop('is_nucleus',
                                     _pdata.is_nucleus(self.pdg_id[0]))
        #: (bool) particle is a hadron
        self.is_hadron = kwargs.pop('is_hadron',
                                    _pdata.is_hadron(self.pdg_id[0]))
        #: (bool) particle is a hadron
        self.is_lepton = kwargs.pop('is_lepton',
                                    _pdata.is_lepton(self.pdg_id[0]))
        #: Mass, charge, neutron number
        self.A, self.Z, self.N = getAZN(self.pdg_id[0])
        #: (float) ctau in cm
        self.ctau = ctau
        #: (float) mass in GeV
        self.mass = mass
        #: (str) species name in string representation
        self.name = name
        #: (bool) particle is stable
        self.is_stable = not self.ctau < np.inf

    def set_cs(self, cs_db):
        """Set cross section adn recalculate the dependent variables"""

        self.current_cross_sections = cs_db.iam
        self.cs = cs_db[self.pdg_id[0]]
        if sum(self.cs) > 0:
            self.can_interact = True
        else:
            self.can_interact = False
        self._critical_energy()
        self._calculate_mixing_energy()

    def set_hadronic_channels(self, hadronic_db, pmanager):
        """Changes the hadronic interaction model.

        Replaces indexing of the yield dictionary from PDG IDs
        with references from partilcle manager.
        """

        self.current_hadronic_model = hadronic_db.iam
        # Collect MCEqParticle references to children
        # instead of PDG ID as index
        if self.pdg_id in hadronic_db.parents and not self.is_tracking:
            self.is_projectile = True
            self.hadr_secondaries = [
                pmanager.pdg2pref[pid]
                for pid in hadronic_db.relations[self.pdg_id]
            ]
            self.hadr_yields = {}
            for s in self.hadr_secondaries:
                self.hadr_yields[s] = hadronic_db.get_matrix(
                    self.pdg_id, s.pdg_id)
        else:
            self.is_projectile = False
            self.hadr_secondaries = []
            self.hadr_yields = {}

    def set_decay_channels(self, decay_db, pmanager):
        """Populates decay channel and energy distributions"""

        if self.is_stable or self.is_tracking:
            # Variables for decays
            self.children = []
            self.decay_dists = {}
            return

        if self.pdg_id not in decay_db.parents:
            raise Exception('Unstable particle without decay distribution:',
                            self.pdg_id, self.name)

        self.children = []
        self.children = [pmanager[d] for d in decay_db.children(self.pdg_id)]
        self.decay_dists = {}
        for c in self.children:
            self.decay_dists[c] = decay_db.get_matrix(self.pdg_id, c.pdg_id)

    def track_decays(self, tracking_particle):
        children_d = {}
        for c in self.children:
            children_d[c.pdg_id] = c
        if tracking_particle.pdg_id not in list(children_d):
            info(
                17, 'Parent particle {0} does not decay into {1}'.format(
                    self.name, tracking_particle.name))
            return False
        # Copy the decay distribution from original PDG
        self.children.append(tracking_particle)
        self.decay_dists[tracking_particle] = self.decay_dists[children_d[
            tracking_particle.pdg_id]]
        return True

    def track_interactions(self, tracking_particle):
        secondaries_d = {}
        for s in self.hadr_secondaries:
            secondaries_d[s.pdg_id] = s
        if tracking_particle.pdg_id not in list(secondaries_d):
            info(
                17, 'Parent particle {0} does not produce {1} at the vertex'.
                format(self.name, tracking_particle.name))
            return False
        # Copy the decay distribution from original PDG
        self.hadr_secondaries.append(tracking_particle)
        self.hadr_yields[tracking_particle] = self.hadr_yields[secondaries_d[
            tracking_particle.pdg_id]]
        return True

    def is_secondary(self, particle_ref):
        """`True` if this projectile and produces particle `particle_ref`."""
        if not isinstance(particle_ref, self.__class__):
            raise Exception('Argument not of MCEqParticle type.')
        return particle_ref in self.hadr_secondaries

    def is_child(self, particle_ref):
        """`True` if this particle decays into `particle_ref`."""
        if not isinstance(particle_ref, self.__class__):
            raise Exception('Argument not of MCEqParticle type.')
        return particle_ref in self.children

    @property
    def hadridx(self):
        """Returns index range where particle behaves as hadron.

        Returns:
          :func:`tuple` (int,int): range on energy grid
        """
        return (self.mix_idx, self._energy_grid.d)

    @property
    def residx(self):
        """Returns index range where particle behaves as resonance.

        Returns:
          :func:`tuple` (int,int): range on energy grid
        """
        return (0, self.mix_idx)

    @property
    def lidx(self):
        """Returns lower index of particle range in state vector.

        Returns:
          (int): lower index in state vector :attr:`MCEqRun.phi`
        """
        return self.mceqidx * self._energy_grid.d

    @property
    def uidx(self):
        """Returns upper index of particle range in state vector.

        Returns:
          (int): upper index in state vector :attr:`MCEqRun.phi`
        """
        return (self.mceqidx + 1) * self._energy_grid.d

    def inverse_decay_length(self, cut=True):
        r"""Returns inverse decay length (or infinity (np.inf), if
        particle is stable), where the air density :math:`\rho` is
        factorized out.

        Args:
          E (float) : energy in laboratory system in GeV
          cut (bool): set to zero in 'resonance' regime
        Returns:
          (float): :math:`\frac{\rho}{\lambda_{dec}}` in 1/cm
        """
        try:
            dlen = self.mass / self.ctau / (self._energy_grid.c + self.mass)
            if cut:
                dlen[0:self.mix_idx] = 0.
            # Correction for bin average, since dec. length is a steep falling
            # function. This factor averages the value over bin length for
            # 10 bins per decade.
            return 0.989 * dlen
        except ZeroDivisionError:
            return np.ones_like(self._energy_grid.d) * np.inf

    def inel_cross_section(self, mbarn=False):
        """Returns inverse interaction length for A_target given by config.

        Returns:
          (float): :math:`\\frac{1}{\\lambda_{int}}` in cm**2/g
        """
        #: unit - :math:`\text{GeV} \cdot \text{fm}`
        GeVfm = 0.19732696312541853
        #: unit - :math:`\text{GeV} \cdot \text{cm}`
        GeVcm = GeVfm * 1e-13
        #: unit - :math:`\text{GeV}^2 \cdot \text{mbarn}`
        GeV2mbarn = 10.0 * GeVfm**2
        #: unit conversion - :math:`\text{mbarn} \to \text{cm}^2`
        mbarn2cm2 = GeV2mbarn / GeVcm**2
        if mbarn:
            return mbarn2cm2 * self.cs

        return self.cs

    def inverse_interaction_length(self):
        """Returns inverse interaction length for A_target given by config.

        Returns:
          (float): :math:`\\frac{1}{\\lambda_{int}}` in cm**2/g
        """

        m_target = config.A_target * 1.672621 * 1e-24  # <A> * m_proton [g]
        return self.cs / m_target

    def _assign_hadr_dist_idx(self, child, projidx, chidx, cmat):
        """Copies a subset, defined between indices ``projidx`` and ``chiidx``
        from the ``hadr_yields`` into ``cmat``

        Args:
          child (int): PDG ID of final state child/secondary particle
          projidx (int,int): tuple containing index range relative
                             to the projectile's energy grid
          dtridx (int,int): tuple containing index range relative
                            to the child's energy grid
          cmat (numpy.array): array reference to the interaction matrix
        """
        cmat[chidx[0]:chidx[1], projidx[0]:projidx[1]] = self.hadr_yields[
            child][chidx[0]:chidx[1], projidx[0]:projidx[1]]

    def _assign_decay_idx(self, child, projidx, chidx, cmat):
        """Copies a subset, defined between indices ``projidx`` and ``chiidx``
        from the ``hadr_yields`` into ``cmat``

        Args:
          child (int): PDG ID of final state child/secondary particle
          projidx (int,int): tuple containing index range relative
                             to the projectile's energy grid
          dtridx (int,int): tuple containing index range relative
                            to the child's energy grid
          cmat (numpy.array): array reference to the interaction matrix
        """
        cmat[chidx[0]:chidx[1], projidx[0]:projidx[1]] = self.decay_dists[
            child][chidx[0]:chidx[1], projidx[0]:projidx[1]]

    def dN_dxlab(self, kin_energy, sec_pdg, verbose=True, **kwargs):
        r"""Returns :math:`dN/dx_{\rm Lab}` for interaction energy close
        to ``kin_energy`` for hadron-air collisions.

        The function respects modifications applied via :func:`_set_mod_pprod`.

        Args:
            kin_energy (float): approximate interaction kin_energy
            prim_pdg (int): PDG ID of projectile
            sec_pdg (int): PDG ID of secondary particle
            verbose (bool): print out the closest enerkin_energygy
        Returns:
            (numpy.array, numpy.array): :math:`x_{\rm Lab}`, :math:`dN/dx_{\rm Lab}`
        """

        eidx = (np.abs(self._energy_grid.c - kin_energy)).argmin()
        en = self._energy_grid.c[eidx]
        info(10, 'Nearest energy, index: ', en, eidx, condition=verbose)

        m = self.hadr_yields[sec_pdg]
        xl_grid = (self._energy_grid.c[:eidx + 1]) / en
        xl_dist = en * xl_grid * m[:eidx +
                                   1, eidx] / self._energy_grid.w[:eidx + 1]

        return xl_grid, xl_dist

    def dNdec_dxlab(self, kin_energy, sec_pdg, verbose=True, **kwargs):
        r"""Returns :math:`dN/dx_{\rm Lab}` for interaction energy close
        to ``kin_energy`` for hadron-air collisions.

        The function respects modifications applied via :func:`_set_mod_pprod`.

        Args:
            kin_energy (float): approximate interaction energy
            prim_pdg (int): PDG ID of projectile
            sec_pdg (int): PDG ID of secondary particle
            verbose (bool): print out the closest energy
        Returns:
            (numpy.array, numpy.array): :math:`x_{\rm Lab}`, :math:`dN/dx_{\rm Lab}`
        """

        eidx = (np.abs(self._energy_grid.c - kin_energy)).argmin()
        en = self._energy_grid.c[eidx]
        info(10, 'Nearest energy, index: ', en, eidx, condition=verbose)

        m = self.decay_dists[sec_pdg]
        xl_grid = (self._energy_grid.c[:eidx + 1]) / en
        xl_dist = en * xl_grid * m[:eidx +
                                   1, eidx] / self._energy_grid.w[:eidx + 1]

        return xl_grid, xl_dist

    def dN_dEkin(self, kin_energy, sec_pdg, verbose=True, **kwargs):
        r"""Returns :math:`dN/dE_{\rm Kin}` in lab frame for an interaction energy
        close to ``kin_energy`` (total) for hadron-air collisions.

        The function respects modifications applied via :func:`_set_mod_pprod`.

        Args:
            kin_energy (float): approximate interaction energy
            prim_pdg (int): PDG ID of projectile
            sec_pdg (int): PDG ID of secondary particle
            verbose (bool): print out the closest energy
        Returns:
            (numpy.array, numpy.array): :math:`x_{\rm Lab}`, :math:`dN/dx_{\rm Lab}`
        """

        eidx = (np.abs(self._energy_grid.c - kin_energy)).argmin()
        en = self._energy_grid.c[eidx]
        info(10, 'Nearest energy, index: ', en, eidx, condition=verbose)

        m = self.hadr_yields[sec_pdg]
        ekin_grid = self._energy_grid.c
        elab_dist = m[:eidx + 1, eidx]  / self._energy_grid.w[eidx]

        return ekin_grid[:eidx + 1], elab_dist

    def dN_dxf(self,
               energy,
               prim_pdg,
               sec_pdg,
               pos_only=True,
               verbose=True,
               **kwargs):
        r"""Returns :math:`dN/dx_{\rm F}` in c.m. for interaction energy close
        to ``energy`` (lab. not kinetic) for hadron-air collisions.

        The function respects modifications applied via :func:`_set_mod_pprod`.

        Args:
            energy (float): approximate interaction lab. energy
            prim_pdg (int): PDG ID of projectile
            sec_pdg (int): PDG ID of secondary particle
            verbose (bool): print out the closest energy

        Returns:
            (numpy.array, numpy.array): :math:`x_{\rm F}`, :math:`dN/dx_{\rm F}`
        """
        if not hasattr(self, '_ptav_sib23c'):
            # Load spline of average pt distribution as a funtion of log(E_lab) from sib23c
            import pickle
            from os.path import join
            self._ptav_sib23c = pickle.load(
                open(join(config.data_dir, 'sibyll23c_aux.ppd'), 'rb'))[0]

        def xF(xL, Elab, ppdg):

            m = {2212: 0.938, 211: 0.139, 321: 0.493}
            mp = m[2212]

            Ecm = np.sqrt(2 * Elab * mp + 2 * mp**2)
            Esec = xL * Elab
            betacm = np.sqrt((Elab - mp) / (Elab + mp))
            gammacm = (Elab + mp) / Ecm
            avpt = self._ptav_sib23c[ppdg](
                np.log(np.sqrt(Elab**2) - m[np.abs(ppdg)]**2))

            xf = 2 * (-betacm * gammacm * Esec + gammacm *
                      np.sqrt(Esec**2 - m[np.abs(ppdg)]**2 - avpt**2)) / Ecm
            dxl_dxf = 1. / (
                2 *
                (-betacm * gammacm * Elab + xL * Elab**2 * gammacm / np.sqrt(
                    (xL * Elab)**2 - m[np.abs(ppdg)]**2 - avpt**2)) / Ecm)

            return xf, dxl_dxf

        eidx = (np.abs(self._energy_grid.c + self.mass - energy)).argmin()
        en = self._energy_grid.c[eidx] + self.mass
        info(2, 'Nearest energy, index: ', en, eidx, condition=verbose)
        m = self.hadr_yields[sec_pdg]
        xl_grid = (self._energy_grid.c[:eidx + 1] + self.mass) / en
        xl_dist = xl_grid * en * m[:eidx + 1, eidx] / np.diag(
            self._energy_grid.w)[:eidx + 1]
        xf_grid, dxl_dxf = xF(xl_grid, en, sec_pdg)
        xf_dist = xl_dist * dxl_dxf

        if pos_only:
            xf_dist = xf_dist[xf_grid >= 0]
            xf_grid = xf_grid[xf_grid >= 0]
            return xf_grid, xf_dist

        return xf_grid, xf_dist

    def _critical_energy(self):
        """Returns critical energy where decay and interaction
        are balanced.

        Approximate value in Air.

        Returns:
          (float): :math:`\\frac{m\\ 6.4 \\text{km}}{c\\tau}` in GeV
        """
        if self.is_stable or self.ctau <= 0.:
            self.E_crit = np.inf
        else:
            self.E_crit = self.mass * 6.4e5 / self.ctau

    def _calculate_mixing_energy(self):
        """Calculates interaction/decay length in Air and decides if
        the particle has resonance and/or hadron behavior.

        Class attributes :attr:`is_mixed`, :attr:`E_mix`, :attr:`mix_idx`,
        :attr:`is_resonance` are calculated here. If the option `no_mixing`
        is set in config.adv_config particle is forced to be a resonance
        or hadron behavior.

        Args:
          e_grid (numpy.array): energy grid of size :attr:`d`
          max_density (float): maximum density on the integration path (largest
                               decay length)
        """

        cross_over = config.hybrid_crossover
        max_density = config.max_density
        no_mix = False

        d = self._energy_grid.d

        inv_intlen = self.inverse_interaction_length()
        inv_declen = self.inverse_decay_length()
        if (not np.any(np.nan_to_num(inv_declen) > 0.)
                or abs(self.pdg_id[0]) in config.adv_set["exclude_from_mixing"]
                or config.adv_set['no_mixing']
                or self.pdg_id[0] in config.adv_set['disable_decays']):
            self.mix_idx = 0
            self.is_mixed = False
            self.is_resonance = False
            return
        if (np.abs(self.pdg_id[0]) in config.adv_set["force_resonance"]
                or (np.all(inv_declen == 0.) and not self.is_lepton)):
            threshold = 0.
        elif np.any(inv_intlen > 0.):
            with np.errstate(divide='ignore'):
                lint = 1. / inv_intlen
            d_tilde = np.zeros(d)
            d_tilde[inv_declen > 0] = 1. / inv_declen[inv_declen > 0]
            # multiply with maximal density encountered along the
            # integration path
            ldec = d_tilde * max_density
            threshold = ldec / lint
        else:
            threshold = np.inf
            no_mix = True

        if np.max(threshold) < cross_over:
            self.mix_idx = d - 1
            self.E_mix = self._energy_grid.c[self.mix_idx]
            self.is_mixed = False
            self.is_resonance = True

        elif np.min(threshold) > cross_over or no_mix:
            self.mix_idx = 0
            self.is_mixed = False
            self.is_resonance = False
        else:
            self.mix_idx = np.where(ldec / lint > cross_over)[0][0]
            self.E_mix = self._energy_grid.c[self.mix_idx]
            self.is_mixed = True
            self.is_resonance = False

    def __eq__(self, other):
        """Checks name for equality"""
        if isinstance(other, MCEqParticle):
            return self.name == other.name
        else:
            return NotImplemented

    def __neq__(self, other):
        """Checks name for equality"""
        if isinstance(other, MCEqParticle):
            return self.name != other.name
        else:
            return NotImplemented

    def __hash__(self):
        """Instruction for comuting the hash"""
        return hash(self.name)

    def __repr__(self):
        a_string = ("""
        {0}:
        is_hadron     : {1}
        is_lepton     : {2}
        is_nucleus    : {3}
        is_stable     : {4}
        is_mixed      : {5}
        is_resonance  : {6}
        is_tracking   : {7}
        is_projectile : {8}
        mceqidx       : {9}
        E_mix         : {10:2.1e} GeV\n""").format(
            self.name, self.is_hadron, self.is_lepton, self.is_nucleus,
            self.is_stable, self.is_mixed, self.is_resonance, self.is_tracking,
            self.is_projectile, self.mceqidx, self.E_mix)
        return a_string


class ParticleManager(object):
    """Database for objects of :class:`MCEqParticle`.

    Authors:
        Anatoli Fedynitch (DESY)
        Jonas Heinze (DESY)
    """
    def __init__(self, pdg_id_list, energy_grid, cs_db, mod_table=None):
        # (dict) Dimension of primary grid
        self._energy_grid = energy_grid
        # Particle index shortcuts
        #: (dict) Converts PDG ID to index in state vector
        self.pdg2mceqidx = {}
        #: (dict) Converts particle name to index in state vector
        self.pname2mceqidx = {}
        #: (dict) Converts PDG ID to reference of
        # :class:`particlemanager.MCEqParticle`
        self.pdg2pref = {}
        #: (dict) Converts particle name to reference of
        #: class:`particlemanager.MCEqParticle`
        self.pname2pref = {}
        #: (dict) Converts MCEq index to reference of
        #: class:`particlemanager.MCEqParticle`
        self.mceqidx2pref = {}
        #: (dict) Converts index in state vector to PDG ID
        self.mceqidx2pdg = {}
        #: (dict) Converts index in state vector to reference
        # of :class:`particlemanager.MCEqParticle`
        self.mceqidx2pname = {}
        # Save setup of tracked particles to reapply the relations
        # when models change
        self.tracking_relations = {}
        #: (int) Total number of species
        self.nspec = 0
        # save currently applied cross section model
        self.current_cross_sections = None
        # save currently applied hadronic model
        self.current_hadronic_model = None
        # Cross section database
        self._cs_db = cs_db
        # Dictionary to save te tracking particle config
        self.tracking_relations = []
        # Save the tracking relations requested by default tracking
        self._tracking_requested = []

        self._init_categories(particle_pdg_list=pdg_id_list)

        self.print_particle_tables(10)

    def set_cross_sections_db(self, cs_db):
        """Sets the inelastic cross section to each interacting particle.

        This applies to most of the hadrons and does not imply that the
        particle becomes a projectile. parents need in addition defined
        hadronic channels.
        """

        info(5, 'Setting cross section particle variables.')
        if self.current_cross_sections == cs_db.iam:
            info(10, 'Same cross section model not applied to particles.')
            return

        for p in self.cascade_particles:
            p.set_cs(cs_db)
        self.current_cross_sections = cs_db.iam
        self._update_particle_tables()

    def set_decay_channels(self, decay_db):
        """Attaches the references to the decay yield tables to
        each unstable particle"""

        info(5, 'Setting decay info for particles.')
        for p in self.all_particles:
            p.set_decay_channels(decay_db, self)

        self._restore_tracking_setup()
        self._update_particle_tables()

    def set_interaction_model(self,
                              cs_db,
                              hadronic_db,
                              updated_parent_list=None,
                              force=False):
        """Attaches the references to the hadronic yield tables to
        each projectile particle"""

        info(5, 'Setting hadronic secondaries for particles.')
        if (self.current_hadronic_model == hadronic_db.iam and 
            not force and updated_parent_list is None):
            info(10, 'Same hadronic model not applied to particles.')
            return
        if updated_parent_list is not None:
            self._init_categories(updated_parent_list)

        for p in self.cascade_particles:
            p.set_cs(cs_db)
            p.set_hadronic_channels(hadronic_db, self)

        self.current_hadronic_model = hadronic_db.iam
        self._update_particle_tables()

    def set_continuous_losses(self, contloss_db):
        """Set continuous losses terms to particles with ionization
        and radiation losses."""

        for p in self.cascade_particles:
            if p.pdg_id in contloss_db:
                p.has_contloss = True
                p.dEdX = contloss_db[p.pdg_id]

    def add_tracking_particle(self,
                              parent_list,
                              child_pdg,
                              alias_name,
                              from_interactions=False):
        """Allows tracking decay and particle production chains.

        Replaces previous ``obs_particle`` function that allowed to track
        only leptons from decays certain particles. This present feature
        removes the special PDG IDs 71XX, 72XX, etc and allows to define
        any channel like::

            $ particleManagerInstance.add_tracking_particle([211], 14, 'pi_numu')

        This will store muon neutrinos from pion decays under the alias 'pi_numu'.
        Multiple parents are allowed::

            $ particleManagerInstance.add_tracking_particle(
                [411, 421, 431], 14, 'D_numu')

        Args:

            alias (str): Name alias under which the result is accessible in get_solution
            parents (list): list of parent particle PDG ID's
            child (int): Child particle
            from_interactions (bool): track particles from interactions
        """
        from copy import copy

        info(10, 'requested for', parent_list, child_pdg, alias_name)

        for p in parent_list:
            if (p, child_pdg, alias_name,
                    from_interactions) in self._tracking_requested:
                continue
            self._tracking_requested.append(
                (p, child_pdg, alias_name, from_interactions))

        # Check if tracking particle with the alias not yet defined
        # and create new one of necessary
        if alias_name in self.pname2pref:
            info(15, 'Re-using tracking particle', alias_name)
            tracking_particle = self.pname2pref[alias_name]
        elif child_pdg not in self.pdg2pref:
            info(15, 'Tracking child not a available',
                 'for this interaction model, skipping.')
            return
        else:
            info(10, 'Creating new tracking particle')
            # Copy all preferences of the original particle
            tracking_particle = copy(self.pdg2pref[child_pdg])
            tracking_particle.is_tracking = True
            tracking_particle.name = alias_name
            # Find a unique PDG ID for the new tracking particle
            # print child_pdg[0], int(copysign(1000000, child_pdg[0]))
            unique_child_pdg = (child_pdg[0] +
                                int(copysign(1000000, child_pdg[0])),
                                tracking_particle.helicity)

            for i in range(100):
                if unique_child_pdg not in list(self.pdg2pref):
                    break
                info(
                    20, '{0}: trying to find unique_pdg ({1}) for {2}'.format(
                        i, tracking_particle.name, unique_child_pdg))
                unique_child_pdg = (unique_child_pdg[0] +
                                    int(copysign(10000, child_pdg[0])),
                                    tracking_particle.helicity)
            tracking_particle.unique_pdg_id = unique_child_pdg

        # Track if attempt to add the tracking particle succeeded at least once
        track_success = False
        # Include antiparticle

        for parent_pdg in list(
                set(parent_list + [(-p, h) for (p, h) in parent_list])):
            if parent_pdg not in self.pdg2pref:
                info(15,
                     'Parent particle {0} does not exist.'.format(parent_pdg))
                continue
            if (parent_pdg, child_pdg, alias_name,
                    from_interactions) in self.tracking_relations:
                info(
                    20, 'Tracking of {0} from {1} already activated.'.format(
                        tracking_particle.name,
                        self.pdg2pref[parent_pdg].name))
                continue

            if not from_interactions:
                track_method = self.pdg2pref[parent_pdg].track_decays
            else:
                track_method = self.pdg2pref[parent_pdg].track_interactions

            # Check if the tracking is successful. If not the particle is not
            # a child of the parent particle
            if track_method(tracking_particle):
                info(
                    15, 'Parent particle {0} tracking scheduled.'.format(
                        parent_pdg))
                self.tracking_relations.append(
                    (parent_pdg, child_pdg, alias_name, from_interactions))
                track_success = True
        if track_success and tracking_particle.name not in list(
                self.pname2pref):
            tracking_particle.mceqidx = np.max(list(self.mceqidx2pref)) + 1
            self.all_particles.append(tracking_particle)
            self.cascade_particles.append(tracking_particle)
            self._update_particle_tables()
            info(
                10, 'tracking particle {0} successfully added.'.format(
                    tracking_particle.name))

    def track_leptons_from(self,
                           parent_pdg_list,
                           prefix,
                           exclude_em=True,
                           from_interactions=False,
                           use_helicities=False):
        """Adds tracking particles for all leptons coming from decays of parents
        in `parent_pdg_list`.
        """

        leptons = [
            p for p in self.all_particles if p.is_lepton
            and not (p.is_em == exclude_em) and not p.is_tracking
        ]

        for lepton in leptons:
            if not use_helicities and lepton.pdg_id[1] != 0:
                continue
            self.add_tracking_particle(parent_pdg_list, lepton.pdg_id,
                                       prefix + lepton.name, from_interactions)

    def _init_categories(self, particle_pdg_list):
        """Determines the list of particles for calculation and
        returns lists of instances of :class:`data.MCEqParticle` .

        The particles which enter this list are those, which have a
        defined index in the SIBYLL 2.3 interaction model. Included are
        most relevant baryons and mesons and some of their high mass states.
        More details about the particles which enter the calculation can
        be found in :mod:`ParticleDataTool`.

        Returns:
          (tuple of lists of :class:`data.MCEqParticle`): (all particles,
          cascade particles, resonances)
        """
        from MCEq.particlemanager import MCEqParticle

        info(5, "Generating particle list.")

        if particle_pdg_list is not None:
            particles = particle_pdg_list
        else:
            from particletools.tables import SibyllParticleTable
            modtab = SibyllParticleTable()
            particles = modtab.baryons + modtab.mesons + modtab.leptons

        # Remove duplicates
        particles = sorted(list(set(particles)))

        # Initialize particle objects
        particle_list = [
            MCEqParticle(pdg, hel, self._energy_grid, self._cs_db)
            for pdg, hel in particles
        ]

        # Sort by critical energy (= int_len ~== dec_length ~ int_cs/tau)
        particle_list.sort(key=lambda x: x.E_crit, reverse=False)

        # Cascade particles will "live" on the grid and have an mceqidx assigned
        self.cascade_particles = [
            p for p in particle_list if not p.is_resonance
        ]

        self.cascade_particles = sorted(self.cascade_particles,
                                        key=lambda p: abs(p.pdg_id[0]))

        # These particles will only exist implicitely and integated out
        self.resonances = [p for p in particle_list if p.is_resonance]

        # Assign an mceqidx (position in state vector) to each explicit particle
        # Resonances will kepp the default mceqidx = -1
        for mceqidx, h in enumerate(self.cascade_particles):
            h.mceqidx = mceqidx

        self.all_particles = self.cascade_particles + self.resonances
        self._update_particle_tables()

    def _update_particle_tables(self):
        """Update internal mapping tables after changes to the particle
        list occur."""

        self.n_cparticles = len(self.cascade_particles)
        self.dim = self._energy_grid.d
        self.dim_states = self._energy_grid.d * self.n_cparticles

        # Clean all dictionaries
        [
            d.clear() for d in [
                self.pdg2mceqidx, self.pname2mceqidx, self.mceqidx2pdg,
                self.mceqidx2pname, self.mceqidx2pref, self.pdg2pref,
                self.pname2pref
            ]
        ]

        for p in self.all_particles:
            self.pdg2mceqidx[p.unique_pdg_id] = p.mceqidx
            self.pname2mceqidx[p.name] = p.mceqidx
            self.mceqidx2pdg[p.mceqidx] = p.unique_pdg_id
            self.mceqidx2pname[p.mceqidx] = p.name
            self.mceqidx2pref[p.mceqidx] = p
            self.pdg2pref[p.unique_pdg_id] = p
            self.pname2pref[p.name] = p

        self.print_particle_tables(20)

    def _restore_tracking_setup(self):
        """Restores the setup of tracking particles after model changes."""

        info(10, 'Restoring tracking particle setup')

        if not self.tracking_relations and config.enable_default_tracking:
            self._init_default_tracking()
            return

        # Clear tracking_relations for this initialization
        self.tracking_relations = []

        for pid, cid, alias, int_dec in self._tracking_requested:
            if pid not in self.pdg2pref:
                info(15, 'Can not restore {0}, since not in particle list.')
                continue
            self.add_tracking_particle([pid], cid, alias, int_dec)

    def _init_default_tracking(self):
        """Add default tracking particles for leptons from pi, K, and mu"""
        # Init default tracking particles
        info(1, 'Initializing default tracking categories (pi, K, mu)')
        self._tracking_requested_by_default = []
        for parents, prefix, with_helicity in [([(211, 0)], 'pi_', True),
                                               ([(321, 0)], 'k_', True),
                                               ([(13, -1),
                                                 (13, 1)], 'mulr_', False),
                                               ([(13, 0)], 'mu_h0_', False),
                                               ([(13, -1), (13, 0),
                                                 (13, 1)], 'mu_', False),
                                               ([(310, 0),
                                                 (130, 0)], 'K0_', False)]:
            self.track_leptons_from(parents,
                                    prefix,
                                    exclude_em=True,
                                    use_helicities=with_helicity)

        # Track prompt leptons
        self.track_leptons_from([
            p.pdg_id for p in self.all_particles if p.ctau < config.prompt_ctau
        ],
                                'prcas_',
                                exclude_em=True,
                                use_helicities=False)
        # Track leptons from interaction vertices (also prompt)
        self.track_leptons_from(
            [p.pdg_id for p in self.all_particles if p.is_projectile],
            'prres_',
            exclude_em=True,
            from_interactions=True,
            use_helicities=False)

        self.track_leptons_from(
            [p.pdg_id for p in self.all_particles if p.is_em],
            'em_',
            exclude_em=True,
            from_interactions=True,
            use_helicities=False)

    def __contains__(self, pdg_id_or_name):
        """Defines the `in` operator to look for particles"""
        if isinstance(pdg_id_or_name, six.integer_types):
            pdg_id_or_name = (pdg_id_or_name, 0)
        elif isinstance(pdg_id_or_name, six.string_types):
            pdg_id_or_name = (_pdata.pdg_id(pdg_id_or_name), 0)
        return pdg_id_or_name in list(self.pdg2pref)

    def __getitem__(self, pdg_id_or_name):
        """Returns reference to particle object."""
        if isinstance(pdg_id_or_name, tuple):
            return self.pdg2pref[pdg_id_or_name]
        elif isinstance(pdg_id_or_name, six.integer_types):
            return self.pdg2pref[(pdg_id_or_name, 0)]
        else:
            return self.pdg2pref[(_pdata.pdg_id(pdg_id_or_name), 0)]

    def keys(self):
        """Returns pdg_ids of all particles"""
        return [p.pdg_id for p in self.all_particles]

    def __repr__(self):
        str_out = ""
        ident = 3 * ' '
        for s in self.all_particles:
            str_out += s.name + '\n' + ident
            str_out += 'PDG id : ' + str(s.pdg_id) + '\n' + ident
            str_out += 'MCEq idx : ' + str(s.mceqidx) + '\n\n'

        return str_out

    def print_particle_tables(self, min_dbg_lev=2):

        info(min_dbg_lev, "Hadrons and stable particles:", no_caller=True)
        print_in_rows(min_dbg_lev, [
            p.name for p in self.all_particles
            if p.is_hadron and not p.is_resonance and not p.is_mixed
        ])

        info(min_dbg_lev, "\nMixed:", no_caller=True)
        print_in_rows(min_dbg_lev,
                      [p.name for p in self.all_particles if p.is_mixed])

        info(min_dbg_lev, "\nResonances:", no_caller=True)
        print_in_rows(min_dbg_lev,
                      [p.name for p in self.all_particles if p.is_resonance])

        info(min_dbg_lev, "\nLeptons:", no_caller=True)
        print_in_rows(min_dbg_lev, [
            p.name
            for p in self.all_particles if p.is_lepton and not p.is_tracking
        ])
        info(min_dbg_lev, "\nTracking:", no_caller=True)
        print_in_rows(min_dbg_lev,
                      [p.name for p in self.all_particles if p.is_tracking])

        info(min_dbg_lev,
             "\nTotal number of species:",
             self.n_cparticles,
             no_caller=True)

        # list particle indices
        if False:
            info(10, "Particle matrix indices:", no_caller=True)
            some_index = 0
            for p in self.cascade_particles:
                for i in range(self._energy_grid.d):
                    info(10, p.name + '_' + str(i), some_index, no_caller=True)
                    some_index += 1



# -*- coding: utf-8 -*-
"""
:mod:`MCEq.charm_models` --- charmed particle production
========================================================

This module includes classes for custom charmed particle
production. Currently only the MRS model is implemented
as the class :class:`MRS_charm`. The abstract class
:class:`CharmModel` guides the implementation of custom
classes.

The :class:`Yields` instantiates derived classes of
:class:`CharmModel` and calls :func:`CharmModel.get_yield_matrix`
when overwriting a model yield file in
:func:`Yields.set_custom_charm_model`.
"""

import numpy as np
from MCEq.core import config
from MCEq.misc import info
from abc import ABCMeta, abstractmethod
from six import with_metaclass

class CharmModel(with_metaclass(ABCMeta)):
    """Abstract class, from which implemeted charm models can inherit.

    Note:
      Do not instantiate this class directly.

    """

    @abstractmethod
    def get_yield_matrix(self, proj, sec):
        """The implementation of this abstract method returns
        the yield matrix spanned over the energy grid of the calculation.

        Args:
           proj (int): PDG ID of the interacting particle (projectile)
           sec (int): PDG ID of the final state charmed meson (secondary)

        Returns:
           np.array: yield matrix

        Raises:
            NotImplementedError:
        """
        raise NotImplementedError("CharmModel::get_yield_matrix(): " +
                                  "Base class called.")


class MRS_charm(CharmModel):
    """Martin-Ryskin-Stasto charm model.

    The model is described in A. D. Martin, M. G. Ryskin,
    and A. M. Stasto, Acta Physica Polonica B 34, 3273 (2003).
    The parameterization of the inclusive :math:`c\\bar{c}`
    cross-section is given in the appendix of the paper.
    This formula provides the behavior of the cross-section,
    while fragmentation functions and certain scales are
    needed to obtain meson and baryon fluxes as a function
    of the kinematic variable :math:`x_F`. At high energies
    and :math:`x_F > 0.05`, where this model is valid,
    :math:`x_F \\approx x=E_c/E_{proj}`.
    Here, these fragmentation functions are used:

        - :math:`D`-mesons :math:`\\frac{4}{3} x`
        - :math:`\\Lambda`-baryons :math:`\\frac{1}{1.47} x`

    The production ratios between the different types of
    :math:`D`-mesons are stored in the attribute :attr:`cs_scales`
    and :attr:`D0_scale`, where :attr:`D0_scale` is the
    :math:`c\\bar{c}` to :math:`D^0` ratio and :attr:`cs_scales`
    stores the production ratios of :math:`D^\\pm/D^0`,
    :math:`D_s/D^0` and :math:`\\Lambda_c/D^0`.

    Since the model employs only perturbartive production of
    charm, the charge conjugates are symmetric, i.e.
    :math:`\\sigma_{D^+} = \\sigma_{D^-}` etc.

    Args:
      e_grid (np.array): energy grid as it is defined in
                         :class:`MCEqRun`.
      csm (np.array): inelastic cross-sections as used in
                      :class:`MCEqRun`.
    """

    #: fractions of cross-section wrt to D0 cross-section
    cs_scales = {421: 1., 411: 0.5, 431: 0.15, 4122: 0.45}
    #: D0 cross-section wrt to the ccbar cross-section
    D0_scale = 1. / 2.1

    #: hadron projectiles, which are allowed to produce charm
    allowed_proj = [2212, -2212, 2112, -2112, 211, -211, 321, -321]

    #: charm secondaries, which are predicted by this model
    allowed_sec = [411, 421, 431, 4122]

    def __init__(self, e_grid, csm):

        # inverted fragmentation functions
        self.lambda_c_frag = lambda xhad: 1 / 1.47 * xhad
        self.d_frag = lambda xhad: 4. / 3. * xhad

        self.e_grid = e_grid
        self.d = e_grid.size
        self.no_prod = np.zeros(self.d**2).reshape(self.d, self.d)
        self.siginel = csm.get_cs(2212, mbarn=True)

    def sigma_cc(self, E):
        """Returns the integrated ccbar cross-section in mb.

        Note:
          Integration is not going over complete phase-space due to
          limitations of the parameterization.
        """
        from scipy.integrate import quad
        E = np.asarray(E)
        if E.size > 1:
            return 2 * np.array(
                [quad(self.dsig_dx, 0.05, 0.6, args=Ei)[0] for Ei in E])
        else:
            return 2 * quad(self.dsig_dx, 0.05, 0.6, args=E)[0]

    def dsig_dx(self, x, E):
        """Returns the Feynman-:math:`x_F` distribution
        of :math:`\\sigma_{c\\bar{c}}` in mb

        Args:
          x (float or np.array): :math:`x_F`
          E (float): center-of-mass energy in GeV

        Returns:
          float: :math:`\\sigma_{c\\bar{c}}` in mb
        """

        x = np.asarray(x)
        E = np.asarray(E)
        beta = 0.05 - 0.016 * np.log(E / 10e4)
        n, A = None, None
        if E < 1e4:
            return 0.
        elif E >= 1e4 and E < 1e8:
            n = 7.6 + 0.025 * np.log(E / 1e4)
            A = 140 + (11. * np.log(E / 1e2))**1.65
        elif E >= 1e8 and E <= 1e11:
            n = 7.6 + 0.012 * np.log(E / 1e4)
            A = 4100. + 245. * np.log(E / 1e8)
        else:
            raise Exception("MRS_charm()::out of range")
        res = np.zeros_like(x)
        ran = (x > 0.01) & (x < 0.7)
        res[ran] = np.array(A * x[ran]**
                            (beta - 1.) * (1 - x[ran]**1.2)**n / 1e3)
        return res

    def D_dist(self, x, E, mes):
        """Returns the Feynman-:math:`x_F` distribution
        of :math:`\\sigma_{D-mesons}` in mb

        Args:
          x (float or np.array): :math:`x_F`
          E (float): center-of-mass energy in GeV
          mes (int): PDG ID of D-meson: :math:`\\pm421, \\pm431, \\pm411`

        Returns:
          float: :math:`\\sigma_{D-mesons}` in mb
        """
        xc = self.d_frag(x)
        return self.dsig_dx(xc, E) * self.D0_scale * self.cs_scales[mes]

    def LambdaC_dist(self, x, E):
        """Returns the Feynman-:math:`x_F` distribution
        of :math:`\\sigma_{\\Lambda_C}` in mb

        Args:
          x (float or np.array): :math:`x_F`
          E (float): center-of-mass energy in GeV
          mes (int): PDG ID of D-meson: :math:`\\pm421, \\pm431, \\pm411`

        Returns:
          float: :math:`\\sigma_{D-mesons}` in mb
        """
        xc = self.lambda_c_frag(x)
        return self.dsig_dx(xc, E) * self.D0_scale * self.cs_scales[4122]

    def get_yield_matrix(self, proj, sec):
        """Returns the yield matrix in proper format for :class:`MCEqRun`.

        Args:
          proj (int): projectile PDG ID :math:`\\pm` [2212, 211, 321]
          sec (int): charmed particle PDG ID :math:`\\pm` [411, 421, 431, 4122]

        Returns:
          np.array: yield matrix if (proj,sec) combination allowed,
                    else zero matrix
        """
        # TODO: Make this function a member of the base class!

        if (proj not in self.allowed_proj) or (
                abs(sec) not in self.allowed_sec):
            return self.no_prod

        self.xdist = None

        if abs(sec) == 4122 and \
            ((np.sign(proj) != np.sign(sec)) or abs(proj) < 1000):
            return self.no_prod
        else:
            self.xdist = lambda e: self.LambdaC_dist(self.e_grid / e, e) / e
        if abs(sec) != 4122:
            self.xdist = lambda e: self.D_dist(self.e_grid / e, e, abs(sec)) / e

        m_out = np.zeros_like(self.no_prod)

        # convert x distribution to E_sec distribution and distribute on the grid
        for i, e in enumerate(self.e_grid):
            m_out[:, i] = self.xdist(e) / self.siginel[i]

        info(3, 'returning matrix for ({0},{1})'.format(proj, sec))

        return m_out

    def test(self):
        """Plots the meson, baryon and charm quark distribution as shown in
        the plot below.

        .. figure:: graphics/MRS_test.png
            :scale: 50 %
            :alt: output of test function

        """
        import matplotlib.pyplot as plt

        xvec = np.linspace(0.0001, 1., 20)

        # Energy for plotting inclusive cross-sections
        eprobe = 1e7

        plt.figure(figsize=(8.5, 4))
        plt.subplot(121)
        plt.semilogy(
            xvec,
            xvec * self.dsig_dx(xvec, eprobe),
            lw=1.5,
            label=r'$c$-quark')
        plt.semilogy(
            xvec,
            xvec * self.D_dist(xvec, eprobe, 421),
            lw=1.5,
            label=r'$D^0$')
        plt.semilogy(
            xvec,
            xvec * self.D_dist(xvec, eprobe, 411),
            lw=1.5,
            label=r'$D^+$')
        plt.semilogy(
            xvec,
            xvec * self.D_dist(xvec, eprobe, 431),
            lw=1.5,
            label=r'$Ds^+$')
        plt.semilogy(
            xvec,
            xvec * self.LambdaC_dist(xvec, 1e4),
            lw=1.5,
            label=r'$\Lambda_C^+$')
        plt.legend()
        plt.xlabel(r'$x_F$')
        plt.ylabel(r'inclusive $\sigma$ [mb]')

        plt.subplot(122)
        evec = np.logspace(4, 11, 100)
        plt.loglog(
            np.sqrt(evec),
            self.sigma_cc(evec),
            lw=1.5,
            label=r'$\sigma_{c\bar{c}}$')
        plt.legend()
        plt.xlabel(r'$\sqrt{s}$ [GeV]')
        plt.ylabel(r'$\sigma_{c\bar{c}}$ [mb]')
        plt.tight_layout()


class WHR_charm(MRS_charm):
    """Logan Wille, Francis Halzen, Hall Reno.

    The approach is the same as in  A. D. Martin, M. G. Ryskin,
    and A. M. Stasto, Acta Physica Polonica B 34, 3273 (2003).
    The parameterization of the inclusive :math:`c\\bar{c}`
    cross-section is replaced by interpolated tables from the
    calculation. Fragmentation functions and certain scales are
    needed to obtain meson and baryon fluxes as a function
    of the kinematic variable :math:`x_F`. At high energies
    and :math:`x_F > 0.05`, where this model is valid,
    :math:`x_F \\approx x=E_c/E_{proj}`.
    Here, these fragmentation functions are used:

        - :math:`D`-mesons :math:`\\frac{4}{3} x`
        - :math:`\\Lambda`-baryons :math:`\\frac{1}{1.47} x`

    The production ratios between the different types of
    :math:`D`-mesons are stored in the attribute :attr:`cs_scales`
    and :attr:`D0_scale`, where :attr:`D0_scale` is the
    :math:`c\\bar{c}` to :math:`D^0` ratio and :attr:`cs_scales`
    stores the production ratios of :math:`D^\\pm/D^0`,
    :math:`D_s/D^0` and :math:`\\Lambda_c/D^0`.

    Since the model employs only perturbartive production of
    charm, the charge conjugates are symmetric, i.e.
    :math:`\\sigma_{D^+} = \\sigma_{D^-}` etc.

    Args:
      e_grid (np.array): energy grid as it is defined in
                         :class:`MCEqRun`.
      csm (np.array): inelastic cross-sections as used in
                      :class:`MCEqRun`.
    """

    def __init__(self, e_grid, csm):
        import pickle

        self.sig_table = pickle.load(open('references/logan_charm.ppl', 'rb'))
        self.e_idcs = {}
        for i, e in enumerate(e_grid):
            self.e_idcs[e] = i

        MRS_charm.__init__(self, e_grid, csm)

    def dsig_dx(self, x, E):
        """Returns the Feynman-:math:`x_F` distribution
        of :math:`\\sigma_{c\\bar{c}}` in mb

        Args:
          x (float or np.array): :math:`x_F`
          E (float): center-of-mass energy in GeV

        Returns:
          float: :math:`\\sigma_{c\\bar{c}}` in mb
        """
        res = self.sig_table[self.e_idcs[E]](x) * 1e-3  #mub -> mb
        res[res < 0] = 0.
        return res



from __future__ import print_function

import mceq_config as config
from MCEq.core import MCEqRun
import crflux.models as pm
import numpy as np

config.debug_level = 1
config.kernel_config = 'numpy'
config.cuda_gpu_id = 0
config.set_mkl_threads(4)

mceq = MCEqRun(
    interaction_model='SIBYLL23C',
    theta_deg=0.,
    primary_model=(pm.HillasGaisser2012, 'H3a'))


def test_config_and_file_download():
    import mceq_config as config
    import os
    # Import of config triggers data download
    assert config.mceq_db_fname in os.listdir(config.data_dir)


def test_some_angles():

    nmu = []
    for theta in [0., 30., 60., 90]:
        mceq.set_theta_deg(theta)
        mceq.solve()
        nmu.append(
            np.sum(
                mceq.get_solution('mu+', mag=0, integrate=True) +
                mceq.get_solution('mu-', mag=0, integrate=True)))
    assert np.allclose(nmu,
                       [0.004801787364145619,
                        0.0036615779742498415,
                        0.0012506912333444566,
                        8.179093198172332e-06],
                       rtol=1e-4)


def test_switch_interaction_models():
    mlist = [
        'DPMJETIII191',
        'DPMJETIII306',
        'QGSJET01C',
        'QGSJETII03',
        'QGSJETII04',
        'SIBYLL21',
        'SIBYLL23',
        'SIBYLL23C',
        'SIBYLL23CPP']
    count_part = []
    for m in mlist:
        mceq.set_interaction_model(m)
        count_part.append(len(mceq._particle_list))
    assert(count_part == [64, 64, 58, 44, 44, 48, 62, 62, 62])


def test_single_primary():
    energies = [1e3, 1e6, 1e9, 5e10]
    nmu, nnumu, nnue = [], [], []
    mceq.set_interaction_model('SIBYLL23C')
    mceq.set_theta_deg(0.)
    for e in energies:
        mceq.set_single_primary_particle(E=e, pdg_id=2212)
        mceq.solve()
        nmu.append(
            np.sum(
                mceq.get_solution('mu+', mag=0, integrate=True) +
                mceq.get_solution('mu-', mag=0, integrate=True)))
        nnumu.append(
            np.sum(
                mceq.get_solution('numu', mag=0, integrate=True) +
                mceq.get_solution('antinumu', mag=0, integrate=True)))
        nnue.append(
            np.sum(
                mceq.get_solution('nue', mag=0, integrate=True) +
                mceq.get_solution('antinue', mag=0, integrate=True)))

    assert(np.allclose([nmu, nnumu, nnue],
                       [[18.567350622641847, 10486.232899632832, 6094178.175780058, 226728040.8471465],
                        [59.08402527783906, 21111.172321878876,
                            10057901.920254719, 346916423.1257473],
                        [20.007070780314464, 5498.879277541484, 2266361.6016466515, 73347318.16002306]], rtol=1e-2))


def test_corsika_atm():

    from MCEq.geometry.density_profiles import CorsikaAtmosphere

    # Depth at surface and density at X=100 g/cm2
    cka_surf_100 = [
        (1036.099233683902, 0.00015623258808300557),
        (1033.8094962133184, 0.00015782685585891685),
        (1055.861981113731, 0.00016209949387937668),
        (1037.9745770942316, 0.00015591979560837204),
        (1011.4568036341923, 0.00014626903051217024),
        (1019.974568696789, 0.0001464549375212421),
        (1019.9764946890782, 0.0001685608228906579)
    ]
    for iatm, (loc, season) in enumerate([
        ("USStd", None),
        ("BK_USStd", None),
        ("Karlsruhe", None),
        ("ANTARES/KM3NeT-ORCA", None),
        ('SouthPole', 'December'),
        ('PL_SouthPole', 'January'),
        ('PL_SouthPole', 'August'),
    ]):

        cka_obj = CorsikaAtmosphere(loc, season)
        assert np.allclose([cka_obj.max_X, 1. /
                            cka_obj.r_X2rho(100.)], cka_surf_100[iatm])


def test_msis_atm():

    from MCEq.geometry.density_profiles import MSIS00Atmosphere
    msis_surf_100 = [
        (1022.6914983678925, 0.00014380042112573175), (1041.2180457811605, 0.00016046129606232836),
        (1044.6608866969684, 0.00016063221634835724), (1046.427667371285, 0.00016041531186210874),
        (1048.6505423154006, 0.00016107650347480857), (1050.6431802896034, 0.00016342084740033518),
        (1050.2145039327452, 0.00016375664772178006), (1033.3640270683418, 0.00015614485659072835),
        (1045.785578319159, 0.00015970449150213374), (1019.9475650272982, 0.000153212909250962),
        (1020.3640351872195, 0.00015221038616604717), (1047.964376368261, 0.00016218804771381842)
    ]
    for iatm, (loc, season) in enumerate([
        ('SouthPole', "January"),
        ('Karlsruhe', "January"),
        ('Geneva', "January"),
        ('Tokyo', "January"),
        ('SanGrasso', "January"),
        ('TelAviv', "January"),
        ('KSC', "January"),
        ('SoudanMine', "January"),
        ('Tsukuba', "January"),
        ('LynnLake', "January"),
        ('PeaceRiver', "January"),
        ('FtSumner', "January")
    ]):

        msis_obj = MSIS00Atmosphere(loc, season)
        assert np.allclose([msis_obj.max_X, 1. /
                            msis_obj.r_X2rho(100.)], msis_surf_100[iatm])



import os
import six
from time import time
import numpy as np
import mceq_config as config
from MCEq.misc import normalize_hadronic_model_name, info
from MCEq.particlemanager import ParticleManager
import MCEq.data


class MCEqRun(object):
    """Main class for handling the calculation.

    This class is the main user interface for the caclulation. It will
    handle initialization and various error/configuration checks. The
    setup has to be accomplished before invoking the integration routine
    is :func:`MCeqRun.solve`. Changes of configuration, such as:

    - interaction model in :meth:`MCEqRun.set_interaction_model`,
    - primary flux in :func:`MCEqRun.set_primary_model`,
    - zenith angle in :func:`MCEqRun.set_theta_deg`,
    - density profile in :func:`MCEqRun.set_density_model`,
    - member particles of the special ``obs_`` group in :func:`MCEqRun.set_obs_particles`,

    can be made on an active instance of this class, while calling
    :func:`MCEqRun.solve` subsequently to calculate the solution
    corresponding to the settings.

    The result can be retrieved by calling :func:`MCEqRun.get_solution`.


    Args:
      interaction_model (string): PDG ID of the particle
      density_model (string,sting,string): model type, location, season
      primary_model (class, param_tuple): classes derived from
        :class:`CRFluxModels.PrimaryFlux` and its parameters as tuple
      theta_deg (float): zenith angle :math:`\\theta` in degrees,
        measured positively from vertical direction
      adv_set (dict): advanced settings, see :mod:`mceq_config`
      obs_ids (list): list of particle name strings. Those lepton decay
        products will be scored in the special ``obs_`` categories
    """

    def __init__(self, interaction_model, primary_model, theta_deg, **kwargs):

        self._mceq_db = MCEq.data.HDF5Backend()

        interaction_model = normalize_hadronic_model_name(interaction_model)

        # Save atmospheric parameters
        self.density_config = kwargs.pop('density_model', config.density_model)
        self.theta_deg = theta_deg

        #: Interface to interaction tables of the HDF5 database
        self._interactions = MCEq.data.Interactions(mceq_hdf_db=self._mceq_db)

        #: handler for cross-section data of type :class:`MCEq.data.HadAirCrossSections`
        self._int_cs = MCEq.data.InteractionCrossSections(
            mceq_hdf_db=self._mceq_db)

        #: handler for cross-section data of type :class:`MCEq.data.HadAirCrossSections`
        self._cont_losses = MCEq.data.ContinuousLosses(mceq_hdf_db=self._mceq_db,
                                                       material=config.dedx_material)

        #: Interface to decay tables of the HDF5 database
        self._decays = MCEq.data.Decays(mceq_hdf_db=self._mceq_db)

        #: Particle manager (initialized/updated in set_interaction_model)
        self.pman = None

        # Particle list to keep track of previously initialized particles
        self._particle_list = None

        # General Matrix dimensions and shortcuts, controlled by
        # grid of yield matrices
        self._energy_grid = self._mceq_db.energy_grid

        # Initialize solution vector
        self._solution = np.zeros(1)
        # Initialize empty state (particle density) vector
        self._phi0 = np.zeros(1)
        # Initialize matrix builder (initialized in set_interaction_model)
        self.matrix_builder = None
        # Save initial condition (primary flux) to restore after dimensional resizing
        self._restore_initial_condition = []

        # Set interaction model and compute grids and matrices
        self.set_interaction_model(interaction_model,
                                   particle_list=kwargs.pop(
                                       'particle_list', None))

        # Default GPU device id for CUDA
        self._cuda_device = kwargs.pop('cuda_gpu_id', config.cuda_gpu_id)

        # Print particle list after tracking particles have been initialized
        self.pman.print_particle_tables(2)

        # Set atmosphere and geometry TODO do not allow empty density model
        # if density_model is not None:
        self.set_density_model(self.density_config)

        # Set initial flux condition
        if primary_model is not None:
            self.set_primary_model(*primary_model)

    @property
    def e_grid(self):
        """Energy grid (bin centers)"""
        return self._energy_grid.c

    @property
    def e_bins(self):
        """Energy grid (bin edges)"""
        return self._energy_grid.b

    @property
    def e_widths(self):
        """Energy grid (bin widths)"""
        return self._energy_grid.w

    @property
    def dim(self):
        """Energy grid (dimension)"""
        return self._energy_grid.d

    @property
    def dim_states(self):
        """Number of cascade particles times dimension of grid
        (dimension of the equation system)"""
        return self.pman.dim_states
    
    def closest_energy(self, kin_energy):
        """Convenience function to obtain the nearest grid energy
        to the `energy` argument, provided as kinetik energy in lab. frame."""
        eidx = (np.abs(self._energy_grid.c - kin_energy)).argmin()
        return self._energy_grid.c[eidx]

    def get_solution(self,
                     particle_name,
                     mag=0.,
                     grid_idx=None,
                     integrate=False,
                     return_as=config.return_as):
        """Retrieves solution of the calculation on the energy grid.

        Some special prefixes are accepted for lepton names:

        - the total flux of muons, muon neutrinos etc. from all sources/mothers
          can be retrieved by the prefix ``total_``, i.e. ``total_numu``
        - the conventional flux of muons, muon neutrinos etc. from all sources
          can be retrieved by the prefix ``conv_``, i.e. ``conv_numu``
        - correspondigly, the flux of leptons which originated from the decay
          of a charged pion carries the prefix ``pi_`` and from a kaon ``k_``
        - conventional leptons originating neither from pion nor from kaon
          decay are collected in a category without any prefix, e.g. ``numu`` or
          ``mu+``

        Args:
          particle_name (str): The name of the particle such, e.g.
            ``total_mu+`` for the total flux spectrum of positive muons or
            ``pr_antinumu`` for the flux spectrum of prompt anti muon neutrinos
          mag (float, optional): 'magnification factor': the solution is
            multiplied by ``sol`` :math:`= \\Phi \\cdot E^{mag}`
          grid_idx (int, optional): if the integrator has been configured to save
            intermediate solutions on a depth grid, then ``grid_idx`` specifies
            the index of the depth grid for which the solution is retrieved. If
            not specified the flux at the surface is returned
          integrate (bool, optional): return averge particle number instead of
          flux (multiply by bin width)

        Returns:
          (numpy.array): flux of particles on energy grid :attr:`e_grid`
        """

        res = np.zeros(self._energy_grid.d)
        ref = self.pman.pname2pref
        sol = None
        if grid_idx is not None and len(self.grid_sol) == 0:
            raise Exception(
                'Solution not has not been computed on grid. Check input.')
        if grid_idx is None:
            sol = np.copy(self._solution)
        elif grid_idx >= len(self.grid_sol) or grid_idx is None:
            sol = self.grid_sol[-1, :]
        else:
            sol = self.grid_sol[grid_idx, :]

        def sum_lr(lep_str, prefix):
            result = np.zeros(self.dim)
            nsuccess = 0
            for ls in lep_str, lep_str + '_l', lep_str + '_r':
                if prefix + ls not in ref:
                    info(
                        15, 'No separate left and right handed particles,',
                        'or, unavailable particle prefix {0}.'.format(prefix +
                                                                      ls))
                    continue
                result += sol[ref[prefix + ls].lidx:ref[prefix + ls].uidx]
                nsuccess += 1
            if nsuccess == 0 and config.excpt_on_missing_particle:
                raise Exception(
                    'Requested particle {0} not found.'.format(particle_name))
            return result

        lep_str = particle_name.split(
            '_')[1] if '_' in particle_name else particle_name

        if particle_name.startswith('total_'):
            # Note: This has changed from previous MCEq versions,
            # since pi_ and k_ prefixes are mere tracking counters
            # and no full particle species anymore

            res = sum_lr(lep_str, prefix='')

        elif particle_name.startswith('conv_'):
            # Note: This changed from previous MCEq versions,
            # conventional is defined as total - prompt
            res = (self.get_solution('total_' + lep_str,
                                     mag=0,
                                     grid_idx=grid_idx,
                                     integrate=False,
                                     return_as='kinetic energy') -
                   self.get_solution('pr_' + lep_str,
                                     mag=0,
                                     grid_idx=grid_idx,
                                     integrate=False,
                                     return_as='kinetic energy'))

        elif particle_name.startswith('pr_'):
            if 'prcas_' + lep_str in ref:
                res += sum_lr(lep_str, prefix='prcas_')
            if 'prres_' + lep_str in ref:
                res += sum_lr(lep_str, prefix='prres_')
            if 'em_' + lep_str in ref:
                res += sum_lr(lep_str, prefix='em_')
        else:
            try:
                res = sum_lr(particle_name, prefix='')
            except KeyError:
                info(10,
                     'Requested particle {0} not found.'.format(particle_name))

        # When returning in Etot, interpolate on different grid
        if return_as == 'total energy':
            etot_bins = self.e_bins + ref[lep_str].mass
            etot_grid = np.sqrt(etot_bins[1:] * etot_bins[:-1])

            if not integrate:
                return etot_grid, res * etot_grid**mag
            else:
                return etot_grid, res * etot_grid**mag * (etot_bins[1:] -
                                                          etot_bins[:-1])

        elif return_as == 'kinetic energy':
            if not integrate:
                return res * self._energy_grid.c**mag
            else:
                return res * self._energy_grid.c**mag * self._energy_grid.w

        elif return_as == 'total momentum':
            ptot_bins = np.sqrt((self.e_bins + ref[lep_str].mass)**2 -
                                ref[lep_str].mass**2)
            ptot_grid = np.sqrt(ptot_bins[1:] * ptot_bins[:-1])
            dEkindp = ptot_grid / np.sqrt(ptot_grid**2 + ref[lep_str].mass**2)
            res *= dEkindp
            if not integrate:
                return ptot_grid, res * ptot_grid**mag
            else:
                return ptot_grid, res * ptot_grid**mag * (ptot_bins[1:] -
                                                          ptot_bins[:-1])

        else:
            raise Exception(
                "Unknown 'return_as' variable choice.",
                'the options are "kinetic energy", "total energy", "total momentum"'
            )

    def set_interaction_model(self,
                              interaction_model,
                              particle_list=None,
                              update_particle_list=True,
                              force=False):
        """Sets interaction model and/or an external charm model for calculation.

        Decay and interaction matrix will be regenerated automatically
        after performing this call.

        Args:
          interaction_model (str): name of interaction model
          charm_model (str, optional): name of charm model
          force (bool): force loading interaction model
        """
        interaction_model = normalize_hadronic_model_name(interaction_model)

        info(1, interaction_model)

        if not force and (self._interactions.iam == interaction_model
                          ) and particle_list != self._particle_list:
            info(2, 'Skip, since current model identical to',
                 interaction_model + '.')
            return

        self._int_cs.load(interaction_model)

        # TODO: simplify this, stuff not needed anymore
        if not update_particle_list and self._particle_list is not None:
            info(10, 'Re-using particle list.')
            self._interactions.load(interaction_model,
                                    parent_list=self._particle_list)
            self.pman.set_interaction_model(self._int_cs, self._interactions)
            self.pman.set_decay_channels(self._decays)
            self.pman.set_continuous_losses(self._cont_losses)

        elif self._particle_list is None:
            info(10, 'New initialization of particle list.')
            # First initialization
            if particle_list is None:
                self._interactions.load(interaction_model)
            else:
                self._interactions.load(interaction_model,
                                        parent_list=particle_list)

            self._decays.load(parent_list=self._interactions.particles)
            self._particle_list = self._interactions.particles + self._decays.particles
            # Create particle database
            self.pman = ParticleManager(self._particle_list, self._energy_grid,
                                        self._int_cs)
            self.pman.set_interaction_model(self._int_cs, self._interactions)
            self.pman.set_decay_channels(self._decays)
            self.pman.set_continuous_losses(self._cont_losses)
            self.matrix_builder = MatrixBuilder(self.pman)

        elif (update_particle_list and particle_list != self._particle_list):
            info(10, 'Updating particle list.')
            # Updated particle list received
            if particle_list is None:
                self._interactions.load(interaction_model)
            else:
                self._interactions.load(interaction_model,
                                        parent_list=particle_list)
            self._decays.load(parent_list=self._interactions.particles)
            self._particle_list = self._interactions.particles + self._decays.particles
            self.pman.set_interaction_model(
                self._int_cs,
                self._interactions,
                updated_parent_list=self._particle_list)
            self.pman.set_decay_channels(self._decays)
            self.pman.set_continuous_losses(self._cont_losses)

        else:
            raise Exception('Should not happen in practice.')

        # Update dimensions if particle dimensions changed
        self._phi0 = np.zeros(self.dim_states)
        self._solution = np.zeros(self.dim_states)

        # Restore insital condition if present
        if len(self._restore_initial_condition) > 0:
            for con in self._restore_initial_condition:
                con[0](*con[1:])

        # initialize matrices
        self.int_m, self.dec_m = self.matrix_builder.construct_matrices(
            skip_decay_matrix=False)

    def set_primary_model(self, mclass, tag):
        """Sets primary flux model.

        This functions is quick and does not require re-generation of
        matrices.

        Args:
          interaction_model (:class:`CRFluxModel.PrimaryFlux`): reference
          to primary model **class**
          tag (tuple): positional argument list for model class
        """

        info(1, mclass.__name__, tag if tag is not None else '')

        # Save primary flux model for restauration after interaction model changes
        self._restore_initial_condition = [
            (self.set_primary_model, mclass, tag)]
        # Initialize primary model object
        self.pmodel = mclass(tag)
        self.get_nucleon_spectrum = np.vectorize(self.pmodel.p_and_n_flux)

        try:
            self.dim_states
        except AttributeError:
            self.finalize_pmodel = True

        # Save initial condition
        minimal_energy = 3.
        if (2212, 0) in self.pman:
            e_tot = self._energy_grid.c + self.pman[(2212, 0)].mass
        else:
            info(
                10,
                'No protons in eqn system, quering primary flux with kinetic energy.'
            )
            e_tot = self._energy_grid.c

        min_idx = np.argmin(np.abs(e_tot - minimal_energy))
        self._phi0 *= 0
        p_top, n_top = self.get_nucleon_spectrum(e_tot[min_idx:])[1:]
        if (2212, 0) in self.pman:
            self._phi0[min_idx + self.pman[(2212, 0)].lidx:self.pman[(
                2212, 0)].uidx] = 1e-4 * p_top
        else:
            info(
                1,
                'Warning protons not part of equation system, can not set primary flux.'
            )

        if (2112, 0) in self.pman and not self.pman[(2112, 0)].is_resonance:
            self._phi0[min_idx + self.pman[(2112, 0)].lidx:self.pman[(
                2112, 0)].uidx] = 1e-4 * n_top
        elif (2212, 0) in self.pman:
            info(2, 'Neutrons not part of equation system,',
                 'substituting initial flux with protons.')
            self._phi0[min_idx + self.pman[(2212, 0)].lidx:self.pman[(
                2212, 0)].uidx] += 1e-4 * n_top

    def set_single_primary_particle(self, E, corsika_id=None, pdg_id=None, append=False):
        """Set type and kinetic energy of a single primary nucleus to
        calculation of particle yields.

        The functions uses the superposition theorem, where the flux of
        a nucleus with mass A and charge Z is modeled by using Z protons
        and A-Z neutrons at energy :math:`E_{nucleon}= E_{nucleus} / A`
        The nucleus type is defined via :math:`\\text{CORSIKA ID} = A*100 + Z`. For
        example iron has the CORSIKA ID 5226.

        Single leptons or hadrons can be defined by specifiying `pdg_id` instead of
        `corsika_id`.

        The `append` argument can be used to compose an initial state with
        multiple particles. If it is `False` the initial condition is reset to zero
        before adding the particle.

        A continuous input energy range is allowed between
        :math:`50*A~ \\text{GeV} < E_\\text{nucleus} < 10^{10}*A \\text{GeV}`.

        Args:
          E (float): kinetic energy of a nucleus in GeV
          corsika_id (int): ID of a nucleus (see text)
          pdg_id (int): PDG ID of a particle
          append (bool): If True, keep previous state and append a new particle.
        """

        from scipy.linalg import solve
        from MCEq.misc import getAZN_corsika, getAZN

        if corsika_id and pdg_id:
            raise Exception('Provide either corsika or PDG ID')

        info(
            2, 'CORSIKA ID {0}, PDG ID {1}, energy {2:5.3g} GeV'.format(
                corsika_id, pdg_id, E))

        if append == False:
            self._restore_initial_condition = [(self.set_single_primary_particle, E,
                                                corsika_id, pdg_id)]
            self._phi0 *= 0.
        else:
            self._restore_initial_condition.append((self.set_single_primary_particle, E,
                                                    corsika_id, pdg_id))
        egrid = self._energy_grid.c
        ebins = self._energy_grid.b
        ewidths = self._energy_grid.w

        if corsika_id:
            n_nucleons, n_protons, n_neutrons = getAZN_corsika(corsika_id)
        elif pdg_id:
            n_nucleons, n_protons, n_neutrons = getAZN(pdg_id)

        En = E / float(n_nucleons) if n_nucleons > 0 else E

        if En < np.min(self._energy_grid.c):
            raise Exception('energy per nucleon too low for primary ' +
                            str(corsika_id))

        info(3, ('superposition: n_protons={0}, n_neutrons={1}, ' +
                 'energy per nucleon={2:5.3g} GeV').format(
                     n_protons, n_neutrons, En))

        cenbin = np.argwhere(En < ebins)[0][0] - 1

        # Equalize the first three moments for 3 normalizations around the central
        # bin
        emat = np.vstack(
            (ewidths[cenbin - 1:cenbin + 2],
             ewidths[cenbin - 1:cenbin + 2] * egrid[cenbin - 1:cenbin + 2],
             ewidths[cenbin - 1:cenbin + 2] * egrid[cenbin - 1:cenbin + 2]**2))

        if n_nucleons == 0:
            # This case handles other exotic projectiles
            b_particle = np.array([1., En, En**2])
            lidx = self.pman[pdg_id].lidx
            self._phi0[lidx + cenbin - 1:lidx + cenbin + 2] += solve(
                emat, b_particle)
            return

        if n_protons > 0:
            b_protons = np.array(
                [n_protons, En * n_protons, En**2 * n_protons])
            p_lidx = self.pman[2212].lidx

            self._phi0[p_lidx + cenbin - 1:p_lidx + cenbin + 2] += solve(
                emat, b_protons)
        if n_neutrons > 0:
            b_neutrons = np.array(
                [n_neutrons, En * n_neutrons, En**2 * n_neutrons])
            n_lidx = self.pman[2112].lidx
            self._phi0[n_lidx + cenbin - 1:n_lidx + cenbin + 2] += solve(
                emat, b_neutrons)

    def set_initial_spectrum(self, spectrum, pdg_id=None, append=False):
        """Set a user-defined spectrum for an arbitrary species as initial condition. 

        This function is an equivalent to :func:`set_single_primary_particle`. It
        allows to define an arbitrary spectrum for each available particle species
        as initial condition for the integration. Set the `append` argument to `True`
        for subsequent species to define initial spectra combined from different particles.

        The (differential) spectrum has to be distributed on the energy grid as dN/dptot, i.e.
        divided by the bin widths and with the total momentum units in GeV(/c).

        Args:
          spectrum (np.array): spectrum dN/dptot
          pdg_id (int): PDG ID in case of a particle 
        """

        from MCEq.misc import getAZN_corsika, getAZN

        info(2, 'PDG ID {0}'.format(pdg_id))

        if not append:
            self._restore_initial_condition = [(self.set_initial_spectrum,
                                                pdg_id, append)]
            self._phi0 *= 0
        else:
            self._restore_initial_condition.append((self.set_initial_spectrum,
                                                    pdg_id, append))
        egrid = self._energy_grid.c
        ebins = self._energy_grid.b
        ewidths = self._energy_grid.w

        if len(spectrum) != self.dim:
            raise Exception(
                'Lengths of spectrum and energy grid do not match.')

        self._phi0[self.pman[pdg_id].lidx:self.pman[pdg_id].uidx] += spectrum

    def set_density_model(self, density_config):
        """Sets model of the atmosphere.

        To choose, for example, a CORSIKA parametrization for the Southpole in January,
        do the following::

            mceq_instance.set_density_model(('CORSIKA', ('PL_SouthPole', 'January')))

        More details about the choices can be found in :mod:`MCEq.geometry.density_profiles`. Calling
        this method will issue a recalculation of the interpolation and the integration path.

        Args:
          density_config (tuple of strings): (parametrization type, arguments)
        """
        import MCEq.geometry.density_profiles as dprof

        base_model, model_config = density_config

        available_models = [
            'MSIS00', 'MSIS00_IC', 'CORSIKA', 'AIRS', 'Isothermal',
            'GeneralizedTarget'
        ]

        if base_model not in available_models:
            info(0, 'Unknown density model. Available choices are:\n',
                 '\n'.join(available_models))
            raise Exception('Choose a different profile.')

        info(1, 'Setting density profile to', base_model, model_config)

        if base_model == 'MSIS00':
            self.density_model = dprof.MSIS00Atmosphere(*model_config)
        elif base_model == 'MSIS00_IC':
            self.density_model = dprof.MSIS00IceCubeCentered(*model_config)
        elif base_model == 'CORSIKA':
            self.density_model = dprof.CorsikaAtmosphere(*model_config)
        elif base_model == 'AIRS':
            self.density_model = dprof.AIRSAtmosphere(*model_config)
        elif base_model == 'Isothermal':
            self.density_model = dprof.IsothermalAtmosphere(*model_config)
        elif base_model == 'GeneralizedTarget':
            self.density_model = dprof.GeneralizedTarget()
        else:
            raise Exception('Unknown atmospheric base model.')
        self.density_config = density_config

        if self.theta_deg is not None and base_model != 'GeneralizedTarget':
            self.set_theta_deg(self.theta_deg)
        elif base_model == 'GeneralizedTarget':
            self.integration_path = None

        # TODO: Make the pman aware of that density might have changed and
        # indices as well
        # self.pmod._gen_list_of_particles()

    def set_theta_deg(self, theta_deg):
        """Sets zenith angle :math:`\\theta` as seen from a detector.

        Currently only 'down-going' angles (0-90 degrees) are supported.

        Args:
          atm_config (tuple of strings): (parametrization type, location string, season string)
        """
        info(2, 'Zenith angle {0:6.2f}'.format(theta_deg))

        if self.density_config[0] == 'GeneralizedTarget':
            raise Exception('GeneralizedTarget does not support angles.')

        if self.density_model.theta_deg == theta_deg:
            info(2,
                 'Theta selection correponds to cached value, skipping calc.')
            return

        self.density_model.set_theta(theta_deg)
        self.integration_path = None

    def set_mod_pprod(self,
                      prim_pdg,
                      sec_pdg,
                      x_func,
                      x_func_args,
                      delay_init=False):
        """Sets combination of projectile/secondary for error propagation.

        The production spectrum of ``sec_pdg`` in interactions of
        ``prim_pdg`` is modified according to the function passed to
        :func:`InteractionYields.init_mod_matrix`

        Args:
          prim_pdg (int): interacting (primary) particle PDG ID
          sec_pdg (int): secondary particle PDG ID
          x_func (object): reference to function
          x_func_args (tuple): arguments passed to ``x_func``
          delay_init (bool): Prevent init of mceq matrices if you are
                             planning to add more modifications
        """
        info(
            1, '{0}/{1}, {2}, {3}'.format(prim_pdg, sec_pdg, x_func.__name__,
                                          str(x_func_args)))

        init = self._interactions._set_mod_pprod(prim_pdg, sec_pdg, x_func,
                                                 x_func_args)

        # Need to regenerate matrices completely
        return int(init)

        # if init and not delay_init:
        #     self.int_m, self.dec_m = self.matrix_builder.construct_matrices(
        #         skip_decay_matrix=True)
        #     return 0

    def unset_mod_pprod(self, dont_fill=False):
        """Removes modifications from :func:`MCEqRun.set_mod_pprod`.

        Args:
          skip_fill (bool): If `true` do not regenerate matrices
          (has to be done at a later step by hand)
        """
        from collections import defaultdict
        info(1, 'Particle production modifications reset to defaults.')

        self._interactions.mod_pprod = defaultdict(lambda: {})
        # Need to regenerate matrices completely
        if not dont_fill:
            self.regenerate_matrices()

    def regenerate_matrices(self, skip_decay_matrix=False):
        """Call this function after applying particle prod. modifications aka
        Barr parameters"""

        # TODO: Not all particles need to be reset and there is some performance loss
        # This can be optmized by refreshing only the particles that change or through
        # lazy evaluation, i.e. hadronic channels dict. calls data.int..get_matrix on demand
        self.pman.set_interaction_model(self._int_cs,
                                        self._interactions,
                                        force=True)
        self.int_m, self.dec_m = self.matrix_builder.construct_matrices(
            skip_decay_matrix=skip_decay_matrix)

    def solve(self, int_grid=None, grid_var='X', **kwargs):
        """Launches the solver.

        The setting `integrator` in the config file decides which solver
        to launch.

        Args:
          int_grid (list): list of depths at which results are recorded
          grid_var (str): Can be depth `X` or something else (currently only `X` supported)
          kwargs (dict): Arguments are passed directly to the solver methods.

        """
        info(2, "Launching {0} solver".format(config.integrator))

        if int_grid is not None and np.any(np.diff(int_grid) < 0):
            raise Exception('The X values in int_grid are required to be strickly',
                            'increasing.')

        # Calculate integration path if not yet happened
        self._calculate_integration_path(int_grid, grid_var)

        phi0 = np.copy(self._phi0)
        nsteps, dX, rho_inv, grid_idcs = self.integration_path

        info(2, 'for {0} integration steps.'.format(nsteps))

        import MCEq.solvers

        start = time()

        if config.kernel_config.lower() == 'numpy':
            kernel = MCEq.solvers.solv_numpy
            args = (nsteps, dX, rho_inv, self.int_m, self.dec_m, phi0,
                    grid_idcs)

        elif (config.kernel_config.lower() == 'cuda'):
            kernel = MCEq.solvers.solv_CUDA_sparse
            try:
                self.cuda_context.set_matrices(self.int_m, self.dec_m)
            except AttributeError:
                from MCEq.solvers import CUDASparseContext
                self.cuda_context = CUDASparseContext(
                    self.int_m, self.dec_m, device_id=self._cuda_device)
            args = (nsteps, dX, rho_inv, self.cuda_context, phi0, grid_idcs)

        elif (config.kernel_config.lower() == 'mkl'):
            kernel = MCEq.solvers.solv_MKL_sparse
            args = (nsteps, dX, rho_inv, self.int_m, self.dec_m, phi0,
                    grid_idcs)

        else:
            raise Exception(
                "Unsupported integrator setting '{0}'.".format(
                    config.kernel_config))

        self._solution, self.grid_sol = kernel(*args)

        info(
            2, 'time elapsed during integration: {0:5.2f}sec'.format(time() -
                                                                     start))

    def _calculate_integration_path(self, int_grid, grid_var, force=False):

        if (self.integration_path and np.alltrue(int_grid == self.int_grid)
                and np.alltrue(self.grid_var == grid_var) and not force):
            info(5, 'skipping calculation.')
            return

        self.int_grid, self.grid_var = int_grid, grid_var
        if grid_var != 'X':
            raise NotImplementedError(
                'the choice of grid variable other than the depth X are not possible, yet.'
            )

        max_X = self.density_model.max_X
        ri = self.density_model.r_X2rho
        max_lint = self.matrix_builder.max_lint
        max_ldec = self.matrix_builder.max_ldec
        info(2, 'X_surface = {0:7.2f}g/cm2'.format(max_X))

        dX_vec = []
        rho_inv_vec = []

        X = 0.0
        step = 0
        grid_step = 0
        grid_idcs = []

        # The factor 0.95 means 5% inbound from stability margin of the
        # Euler intergrator.
        if (max_ldec / config.max_density > max_lint
                and config.leading_process == 'decays'):
            info(3, "using decays as leading eigenvalues")
            def delta_X(X): return config.stability_margin / (max_ldec * ri(X))
        else:
            info(2, "using interactions as leading eigenvalues")
            def delta_X(X): return config.stability_margin / max_lint

        dXmax = config.dXmax
        while X < max_X:
            dX = min(delta_X(X), dXmax)
            if (np.any(int_grid) and (grid_step < len(int_grid))
                    and (X + dX >= int_grid[grid_step])):
                dX = int_grid[grid_step] - X
                grid_idcs.append(step)
                grid_step += 1
            dX_vec.append(dX)
            rho_inv_vec.append(ri(X))
            X = X + dX
            step += 1

        # Integrate
        dX_vec = np.array(dX_vec)
        rho_inv_vec = np.array(rho_inv_vec)

        self.integration_path = len(dX_vec), dX_vec, rho_inv_vec, grid_idcs

    def n_particles(self, label, grid_idx=None, min_energy_cutoff=1e-1):
        """Returns number of particles of type `label` at a grid step above
        an energy threshold for counting."""
        ie_min = np.argmin(
            np.abs(self.e_bins -
                   self.e_bins[self.e_bins >= min_energy_cutoff][0]))
        info(
            10,
            'Energy cutoff for particle number calculation {0:4.3e} GeV'.format(
                self.e_bins[ie_min]))
        info(
            15,
            'First bin is between {0:3.2e} and {1:3.2e} with midpoint {2:3.2e}'
            .format(self.e_bins[ie_min], self.e_bins[ie_min + 1],
                    self.e_grid[ie_min]))
        return np.sum(
            self.get_solution(label, mag=0, integrate=True, grid_idx=grid_idx)[ie_min:])

    def n_mu(self, grid_idx=None, min_energy_cutoff=1e-1):
        """Returns muon number at a grid step above
        an energy threshold for counting."""
        return (self.n_particles('total_mu+', grid_idx=grid_idx, min_energy_cutoff=min_energy_cutoff) +
                self.n_particles('total_mu-', grid_idx=grid_idx, min_energy_cutoff=min_energy_cutoff))

    def n_e(self, grid_idx=None, min_energy_cutoff=1e-1):
        """Returns muon number at a grid step above
        an energy threshold for counting."""
        return (self.n_particles('e+', grid_idx=grid_idx, min_energy_cutoff=min_energy_cutoff) +
                self.n_particles('e-', grid_idx=grid_idx, min_energy_cutoff=min_energy_cutoff))

    def z_factor(self, projectile_pdg, secondary_pdg, definition='primary_e'):
        """Energy dependent Z-factor according to Thunman et al. (1996)"""

        proj = self.pman[projectile_pdg]
        sec = self.pman[secondary_pdg]

        if not proj.is_projectile:
            raise Exception('{0} is not a projectile particle.'.format(
                proj.name))
        info(
            10, 'Computing e-dependent Zfactor for {0} -> {1}'.format(
                proj.name, sec.name))
        if not proj.is_secondary(sec):
            raise Exception('{0} is not a secondary particle of {1}.'.format(
                sec.name, proj.name))

        if proj == 2112:
            nuc_flux = self.pmodel.p_and_n_flux(self.e_grid)[2]
        else:
            nuc_flux = self.pmodel.p_and_n_flux(self.e_grid)[1]
        zfac = np.zeros(self.dim)

        smat = proj.hadr_yields[sec]
        proj_cs = proj.inel_cross_section()
        zfac = np.zeros_like(self.e_grid)

        # Definition wrt CR energy (different from Thunman) on x-axis
        if definition == 'primary_e':
            min_energy = 2.
            for p_eidx, e in enumerate(self.e_grid):
                if e < min_energy:
                    min_idx = p_eidx + 1
                    continue
                zfac[p_eidx] = np.sum(
                    smat[min_idx:p_eidx + 1, p_eidx] * nuc_flux[p_eidx] /
                    nuc_flux[min_idx:p_eidx + 1] * proj_cs[p_eidx] /
                    proj_cs[min_idx:p_eidx + 1])
            return zfac
        else:
            # Like in Thunman et al. 1996
            for p_eidx, _ in enumerate(self.e_grid):
                zfac[p_eidx] = np.sum(smat[p_eidx, p_eidx:] *
                                      nuc_flux[p_eidx:] / nuc_flux[p_eidx] *
                                      proj_cs[p_eidx:] / proj_cs[p_eidx])
            return zfac

    def decay_z_factor(self, parent_pdg, child_pdg):
        """Energy dependent Z-factor according to Lipari (1993)."""

        proj = self.pman[parent_pdg]
        sec = self.pman[child_pdg]

        if proj.is_stable:
            raise Exception('{0} does not decay.'.format(proj.name))
        info(
            10, 'Computing e-dependent decay Zfactor for {0} -> {1}'.format(
                proj.name, sec.name))
        if not proj.is_child(sec):
            raise Exception('{0} is not a a child particle of {1}.'.format(
                sec.name, proj.name))

        cr_gamma = self.pmodel.nucleon_gamma(self.e_grid)
        zfac = np.zeros(self.dim)

        zfac = np.zeros_like(self.e_grid)
        for p_eidx, e in enumerate(self.e_grid):
            # if e < min_energy:
            #     min_idx = p_eidx + 1
            #     continue
            xlab, xdist = proj.dNdec_dxlab(e, sec)
            zfac[p_eidx] = np.trapz(xlab**(-cr_gamma[p_eidx] - 2.) * xdist,
                                    x=xlab)
        return zfac


class MatrixBuilder(object):
    """This class constructs the interaction and decay matrices."""

    def __init__(self, particle_manager):
        self._pman = particle_manager
        self._energy_grid = self._pman._energy_grid
        self.int_m = None
        self.dec_m = None

        self._construct_differential_operator()

    def construct_matrices(self, skip_decay_matrix=False):
        r"""Constructs the matrices for calculation.

        These are:

        - :math:`\boldsymbol{M}_{int} = (-\boldsymbol{1} + \boldsymbol{C}){\boldsymbol{\Lambda}}_{int}`,
        - :math:`\boldsymbol{M}_{dec} = (-\boldsymbol{1} + \boldsymbol{D}){\boldsymbol{\Lambda}}_{dec}`.

        For debug_levels >= 2 some general information about matrix shape and the number of
        non-zero elements is printed. The intermediate matrices :math:`\boldsymbol{C}` and
        :math:`\boldsymbol{D}` are deleted afterwards to save memory.

        Set the ``skip_decay_matrix`` flag to avoid recreating the decay matrix. This is not necessary
        if, for example, particle production is modified, or the interaction model is changed.

        Args:
          skip_decay_matrix (bool): Omit re-creating D matrix

        """

        from itertools import product
        info(
            3, "Start filling matrices. Skip_decay_matrix = {0}".format(
                skip_decay_matrix))

        self._fill_matrices(skip_decay_matrix=skip_decay_matrix)

        cparts = self._pman.cascade_particles

        # interaction part
        # -I + C
        # In first interaction mode it is just C
        self.max_lint = 0.

        for parent, child in product(cparts, cparts):
            idx = (child.mceqidx, parent.mceqidx)
            # Main diagonal
            if child.mceqidx == parent.mceqidx and parent.can_interact:
                # Substract unity from the main diagonals
                info(10, 'substracting main C diagonal from', child.name,
                     parent.name)
                self.C_blocks[idx][np.diag_indices(self.dim)] -= 1.

            if idx in self.C_blocks:
                # Multiply with Lambda_int and keep track the maximal
                # interaction length for the calculation of integration steps
                self.max_lint = np.max([
                    self.max_lint,
                    np.max(parent.inverse_interaction_length())
                ])
                self.C_blocks[idx] *= parent.inverse_interaction_length()

            if child.mceqidx == parent.mceqidx and parent.has_contloss:
                if config.enable_muon_energy_loss and abs(
                        parent.pdg_id[0]) == 13:
                    info(5, 'Cont. loss for', parent.name)
                    self.C_blocks[idx] += self.cont_loss_operator(
                        parent.pdg_id)
                if config.enable_em_ion and abs(parent.pdg_id[0]) == 11:
                    info(5, 'Cont. loss for', parent.name)
                    self.C_blocks[idx] += self.cont_loss_operator(
                        parent.pdg_id)

        self.int_m = self._csr_from_blocks(self.C_blocks)
        # -I + D

        if not skip_decay_matrix or self.dec_m is None:
            self.max_ldec = 0.
            for parent, child in product(cparts, cparts):
                idx = (child.mceqidx, parent.mceqidx)
                # Main diagonal
                if child.mceqidx == parent.mceqidx and not parent.is_stable:
                    # Substract unity from the main diagonals
                    info(10, 'substracting main D diagonal from', child.name,
                         parent.name)
                    self.D_blocks[idx][np.diag_indices(self.dim)] -= 1.
                if idx not in self.D_blocks:
                    info(25, parent.pdg_id[0], child.pdg_id, 'not in D_blocks')
                    continue
                # Multiply with Lambda_dec and keep track of the
                # maximal decay length for the calculation of integration steps
                self.max_ldec = max(
                    [self.max_ldec,
                     np.max(parent.inverse_decay_length())])
                self.D_blocks[idx] *= parent.inverse_decay_length()

            self.dec_m = self._csr_from_blocks(self.D_blocks)

        for mname, mat in [('C', self.int_m), ('D', self.dec_m)]:
            mat_density = (float(mat.nnz) / float(np.prod(mat.shape)))
            info(5, "{0} Matrix info:".format(mname))
            info(5, "    density    : {0:3.2%}".format(mat_density))
            info(5, "    shape      : {0} x {1}".format(*mat.shape))
            info(5, "    nnz        : {0}".format(mat.nnz))
            info(10, "    sum        :", mat.sum())

        info(3, "Done filling matrices.")

        return self.int_m, self.dec_m

    def _average_operator(self, op_mat):
        """Averages the continuous loss operator by performing
        1/max_step explicit euler steps"""

        n_steps = int(1. / config.loss_step_for_average)
        info(
            10,
            'Averaging continuous loss using {0} intermediate steps.'.format(
                n_steps))

        op_step = np.eye(
            self._energy_grid.d) + op_mat * config.loss_step_for_average
        return np.linalg.matrix_power(op_step, n_steps) - np.eye(
            self._energy_grid.d)

    def cont_loss_operator(self, pdg_id):
        """Returns continuous loss operator that can be summed with appropriate
        position in the C matrix."""
        op_mat = -np.diag(1 / self._energy_grid.c).dot(
            self.op_matrix.dot(np.diag(self._pman[pdg_id].dEdX)))

        if config.average_loss_operator:
            return self._average_operator(op_mat)
        else:
            return op_mat

    @property
    def dim(self):
        """Energy grid (dimension)"""
        return self._pman.dim

    @property
    def dim_states(self):
        """Number of cascade particles times dimension of grid
        (dimension of the equation system)"""
        return self._pman.dim_states

    def _zero_mat(self):
        """Returns a new square zero valued matrix with dimensions of grid.
        """
        return np.zeros((self._pman.dim, self._pman.dim))

    def _csr_from_blocks(self, blocks):
        """Construct a csr matrix from a dictionary of submatrices (blocks)

        Note::

            It's super pain the a** to construct a properly indexed sparse matrix
            directly from the blocks, since bmat totally messes up the order.
        """
        from scipy.sparse import csr_matrix

        new_mat = np.zeros((self.dim_states, self.dim_states))
        for (c, p), d in six.iteritems(blocks):
            rc, rp = self._pman.mceqidx2pref[c], self._pman.mceqidx2pref[p]
            try:
                new_mat[rc.lidx:rc.uidx, rp.lidx:rp.uidx] = d
            except ValueError:
                raise Exception(
                    'Dimension mismatch: matrix {0}x{1}, p={2}:({3},{4}), c={5}:({6},{7})'
                    .format(self.dim_states, self.dim_states, rp.name, rp.lidx,
                            rp.uidx, rc.name, rc.lidx, rc.uidx))
        return csr_matrix(new_mat)

    def _follow_chains(self, p, pprod_mat, p_orig, idcs, propmat, reclev=0):
        """Some recursive magic.
        """
        info(40, reclev * '\t', 'entering with', p.name)
        # print 'orig, p', p_orig.pdg_id, p.pdg_id
        for d in p.children:
            info(40, reclev * '\t', 'following to', d.name)
            if not d.is_resonance:
                # print 'adding stuff', p_orig.pdg_id, p.pdg_id, d.pdg_id
                dprop = self._zero_mat()
                p._assign_decay_idx(d, idcs, d.hadridx, dprop)
                propmat[(d.mceqidx, p_orig.mceqidx)] += dprop.dot(pprod_mat)

            if config.debug_level >= 20:
                pstr = 'res'
                dstr = 'Mchain'
                if idcs == p.hadridx:
                    pstr = 'prop'
                    dstr = 'Mprop'
                info(
                    40, reclev * '\t',
                    'setting {0}[({1},{3})->({2},{4})]'.format(
                        dstr, p_orig.name, d.name, pstr, 'prop'))

            if d.is_mixed or d.is_resonance:
                dres = self._zero_mat()
                p._assign_decay_idx(d, idcs, d.residx, dres)
                reclev += 1
                self._follow_chains(d, dres.dot(pprod_mat), p_orig, d.residx,
                                    propmat, reclev)
            else:
                info(20, reclev * '\t', '\t terminating at', d.name)

    def _fill_matrices(self, skip_decay_matrix=False):
        """Generates the interaction and decay matrices from scratch.
        """
        from collections import defaultdict

        # Fill decay matrix blocks
        if not skip_decay_matrix or self.dec_m is None:
            # Initialize empty D matrix
            self.D_blocks = defaultdict(lambda: self._zero_mat())
            for p in self._pman.cascade_particles:
                # Fill parts of the D matrix related to p as mother
                if not p.is_stable and bool(p.children) and not p.is_tracking:
                    self._follow_chains(p,
                                        np.diag(np.ones((self.dim))),
                                        p,
                                        p.hadridx,
                                        self.D_blocks,
                                        reclev=0)
                else:
                    info(20, p.name, 'stable or not added to D matrix')

        # Initialize empty C blocks
        self.C_blocks = defaultdict(lambda: self._zero_mat())
        for p in self._pman.cascade_particles:
            # if p doesn't interact, skip interaction matrices
            if not p.is_projectile:
                if p.is_hadron:
                    info(
                        1, 'No interactions by {0} ({1}).'.format(
                            p.name, p.pdg_id))
                continue
            for s in p.hadr_secondaries:
                # if s not in self.pman.cascade_particles:
                #     print 'Doing nothing with', p.pdg_id, s.pdg_id
                #     continue

                if not s.is_resonance:
                    cmat = self._zero_mat()
                    p._assign_hadr_dist_idx(s, p.hadridx, s.hadridx, cmat)
                    self.C_blocks[(s.mceqidx, p.mceqidx)] += cmat

                cmat = self._zero_mat()
                p._assign_hadr_dist_idx(s, p.hadridx, s.residx, cmat)
                self._follow_chains(s,
                                    cmat,
                                    p,
                                    s.residx,
                                    self.C_blocks,
                                    reclev=1)

    def _construct_differential_operator(self):
        """Constructs a derivative operator for the contiuous losses.

        This implmentation uses a 6th-order finite differences operator,
        only depends on the energy grid. This is an operator for a sub-matrix
        of dimension (energy grid, energy grid) for a single particle. It
        can be likewise applied to all particle species. The dEdX values are
        applied later in ...
        """
        # First rows of operator matrix (values are truncated at the edges
        # of a matrix.)
        diags_leftmost = [0, 1, 2, 3]
        coeffs_leftmost = [-11, 18, -9, 2]
        denom_leftmost = 6
        diags_left_1 = [-1, 0, 1, 2, 3]
        coeffs_left_1 = [-3, -10, 18, -6, 1]
        denom_left_1 = 12
        diags_left_2 = [-2, -1, 0, 1, 2, 3]
        coeffs_left_2 = [3, -30, -20, 60, -15, 2]
        denom_left_2 = 60

        # Centered diagonals
        # diags = [-3, -2, -1, 1, 2, 3]
        # coeffs = [-1, 9, -45, 45, -9, 1]
        # denom = 60.
        diags = diags_left_2
        coeffs = coeffs_left_2
        denom = 60.

        # Last rows at the right of operator matrix
        diags_right_2 = [-d for d in diags_left_2[::-1]]
        coeffs_right_2 = [-d for d in coeffs_left_2[::-1]]
        denom_right_2 = denom_left_2
        diags_right_1 = [-d for d in diags_left_1[::-1]]
        coeffs_right_1 = [-d for d in coeffs_left_1[::-1]]
        denom_right_1 = denom_left_1
        diags_rightmost = [-d for d in diags_leftmost[::-1]]
        coeffs_rightmost = [-d for d in coeffs_leftmost[::-1]]
        denom_rightmost = denom_leftmost

        h = np.log(self._energy_grid.b[1:] / self._energy_grid.b[:-1])
        dim_e = self._energy_grid.d
        last = dim_e - 1

        op_matrix = np.zeros((dim_e, dim_e))
        op_matrix[0, np.asarray(diags_leftmost)] = np.asarray(
            coeffs_leftmost) / (denom_leftmost * h[0])
        op_matrix[1, 1 +
                  np.asarray(diags_left_1)] = np.asarray(coeffs_left_1) / (
                      denom_left_1 * h[1])
        op_matrix[2, 2 +
                  np.asarray(diags_left_2)] = np.asarray(coeffs_left_2) / (
                      denom_left_2 * h[2])
        op_matrix[last, last + np.asarray(diags_rightmost)] = np.asarray(
            coeffs_rightmost) / (denom_rightmost * h[last])
        op_matrix[last - 1, last - 1 +
                  np.asarray(diags_right_1)] = np.asarray(coeffs_right_1) / (
                      denom_right_1 * h[last - 1])
        op_matrix[last - 2, last - 2 +
                  np.asarray(diags_right_2)] = np.asarray(coeffs_right_2) / (
                      denom_right_2 * h[last - 2])
        for row in range(3, dim_e - 3):
            op_matrix[row, row +
                      np.asarray(diags)] = np.asarray(coeffs) / (denom *
                                                                 h[row])

        self.op_matrix = op_matrix




from __future__ import print_function
from collections import namedtuple
import numpy as np
import mceq_config as config

#: Energy grid (centers, bind widths, dimension)
energy_grid = namedtuple("energy_grid", ("c", "b", "w", "d"))

#: Matrix with x_lab=E_child/E_parent values
_xmat = None

def normalize_hadronic_model_name(name):
    import re
    """Converts hadronic model name into standard form"""
    return re.sub('[-.]', '', name).upper()


def theta_deg(cos_theta):
    """Converts :math:`\\cos{\\theta}` to :math:`\\theta` in degrees.
    """
    return np.rad2deg(np.arccos(cos_theta))


def theta_rad(theta):
    """Converts :math:`\\theta` from rad to degrees.
    """
    return np.deg2rad(theta)


def gen_xmat(energy_grid):
    """Generates x_lab matrix for a given energy grid"""
    global _xmat
    dims = (energy_grid.d, energy_grid.d)
    if _xmat is None or _xmat.shape != dims:
        _xmat = np.zeros(dims)
        for eidx in range(energy_grid.d):
            xvec = energy_grid.c[:eidx + 1] / energy_grid.c[eidx]
            _xmat[:eidx + 1, eidx] = xvec
    return _xmat


def print_in_rows(min_dbg_level, str_list, n_cols=5):
    """Prints contents of a list in rows `n_cols`
    entries per row.
    """
    if min_dbg_level > config.debug_level:
        return

    ls = len(str_list)
    n_full_length = int(ls / n_cols)
    n_rest = ls % n_cols
    print_str = '\n'
    for i in range(n_full_length):
        print_str += ('"{:}", ' * n_cols).format(*str_list[i * n_cols:(i + 1) *
                                                           n_cols]) + '\n'
    print_str += ('"{:}", ' * n_rest).format(*str_list[-n_rest:])

    print(print_str.strip()[:-1])


def is_charm_pdgid(pdgid):
    """Returns True if particle ID belongs to a heavy (charm) hadron."""

    return ((abs(pdgid) > 400 and abs(pdgid) < 500)
            or (abs(pdgid) > 4000 and abs(pdgid) < 5000))


def _get_closest(value, in_list):
    """Returns the closes value to 'value' from given list."""

    minindex = np.argmin(np.abs(in_list - value * np.ones(len(in_list))))
    return minindex, in_list[minindex]


def getAZN(pdg_id):
    """Returns mass number :math:`A`, charge :math:`Z` and neutron
    number :math:`N` of ``pdg_id``.

    Note::

        PDG ID for nuclei is coded according to 10LZZZAAAI.
        For iron-52 it is 1000260520.

    Args:
        pdgid (int): PDG ID of nucleus/mass group
    Returns:
        (int,int,int): (Z,A) tuple
    """
    Z, A = 1, 1
    if pdg_id < 2000:
        return 0, 0, 0
    elif pdg_id == 2112:
        return 1, 0, 1
    elif pdg_id == 2212:
        return 1, 1, 0
    elif pdg_id > 1000000000:
        A = pdg_id % 1000 / 10
        Z = pdg_id % 1000000 / 10000
        return A, Z, A - Z
    else:
        return 1, 0, 0


def getAZN_corsika(corsikaid):
    """Returns mass number :math:`A`, charge :math:`Z` and neutron
    number :math:`N` of ``corsikaid``.

    Args:
        corsikaid (int): corsika id of nucleus/mass group
    Returns:
        (int,int,int): (Z,A) tuple
    """
    Z, A = 1, 1

    if corsikaid == 14:
        return getAZN(2212)
    if corsikaid >= 100:
        Z = corsikaid % 100
        A = (corsikaid - Z) / 100
    else:
        Z, A = 0, 0

    return A, Z, A - Z


def corsikaid2pdg(corsika_id):
    """Conversion of CORSIKA nuclear code to PDG nuclear code"""
    if corsika_id in [101, 14]:
        return 2212
    elif corsika_id in [100, 13]:
        return 2112
    else:
        A, Z, _ = getAZN_corsika(corsika_id)
        # 10LZZZAAAI
        pdg_id = 1000000000
        pdg_id += 10 * A
        pdg_id += 10000 * Z
        return pdg_id


def pdg2corsikaid(pdg_id):
    """Conversion from nuclear PDG ID to CORSIKA ID.

    Note::

        PDG ID for nuclei is coded according to 10LZZZAAAI.
        For iron-52 it is 1000260520.
    """
    if pdg_id == 2212:
        return 14

    A = pdg_id % 1000 / 10
    Z = pdg_id % 1000000 / 10000

    return A * 100 + Z


def caller_name(skip=2):
    """Get a name of a caller in the format module.class.method

    `skip` specifies how many levels of stack to skip while getting caller
    name. skip=1 means "who calls me", skip=2 "who calls my caller" etc.
    An empty string is returned if skipped levels exceed stack height.abs

    From https://gist.github.com/techtonik/2151727
    """
    import inspect

    stack = inspect.stack()
    start = 0 + skip

    if len(stack) < start + 1:
        return ''

    parentframe = stack[start][0]

    name = []

    if config.print_module:
        module = inspect.getmodule(parentframe)
        # `modname` can be None when frame is executed directly in console
        if module:
            name.append(module.__name__ + '.')

    # detect classname
    if 'self' in parentframe.f_locals:
        # I don't know any way to detect call from the object method
        # there seems to be no way to detect static method call - it will
        # be just a function call

        name.append(parentframe.f_locals['self'].__class__.__name__ + '::')

    codename = parentframe.f_code.co_name
    if codename != '<module>':  # top level usually
        name.append(codename + '(): ')  # function or a method
    else:
        name.append(': ')  # If called from module scope

    del parentframe
    return "".join(name)


def info(min_dbg_level, *message, **kwargs):
    """Print to console if `min_debug_level <= config.debug_level`

    The fuction determines automatically the name of caller and appends
    the message to it. Message can be a tuple of strings or objects
    which can be converted to string using `str()`.

    Args:
        min_dbg_level (int): Minimum debug level in config for printing
        message (tuple): Any argument or list of arguments that casts to str
        condition (bool): Print only if condition is True
        blank_caller (bool): blank the caller name (for multiline output)
        no_caller (bool): don't print the name of the caller

    Authors:
        Anatoli Fedynitch (DESY)
        Jonas Heinze (DESY)
    """
    condition = kwargs.pop('condition', True)
    blank_caller = kwargs.pop('blank_caller', False)
    no_caller = kwargs.pop('no_caller', False)
    if config.override_debug_fcn and min_dbg_level < config.override_max_level:
        fcn_name = caller_name(skip=2).split('::')[-1].split('():')[0]
        if fcn_name in config.override_debug_fcn:
            min_dbg_level = 0

    if condition and min_dbg_level <= config.debug_level:
        message = [str(m) for m in message]
        cname = caller_name() if not no_caller else ''
        if blank_caller:
            cname = len(cname) * ' '
        print(cname + " ".join(message))
